import logging
import re
from collections import OrderedDict
from typing import Optional, List, Set, Tuple, Any

import ujson
from IPy import IP
from ordered_set import OrderedSet
from pytz import timezone

from irrd import __version__
from irrd.conf import get_setting, RPKI_IRR_PSEUDO_SOURCE
from irrd.mirroring.nrtm_generator import NRTMGenerator, NRTMGeneratorException
from irrd.rpki.status import RPKIStatus
from irrd.rpsl.rpsl_objects import (OBJECT_CLASS_MAPPING, lookup_field_names,
                                    RPKI_RELEVANT_OBJECT_CLASSES)
from irrd.scopefilter.status import ScopeFilterStatus
from irrd.storage.database_handler import DatabaseHandler, is_serial_synchronised
from irrd.storage.preload import Preloader
from irrd.storage.queries import RPSLDatabaseQuery, DatabaseStatusQuery
from irrd.utils.validators import parse_as_number, ValidationError
from .query_response import WhoisQueryResponseType, WhoisQueryResponseMode, WhoisQueryResponse
from ..access_check import is_client_permitted

logger = logging.getLogger(__name__)


class WhoisQueryParserException(ValueError):
    """
    This exception can be thrown by parsers of individual commands, and
    is handled in WhoisQueryParser.handle_query to be turned into a
    WhoisQueryResponse with a suitable error message.
    """
    pass


class WhoisQueryParser:
    """
    Parser for all whois-style queries.

    This parser distinguishes RIPE-style, e.g. '-K 192.0.2.1' or '-i mnt-by FOO'
    from IRRD-style, e.g. '!oFOO'.

    Some query flags, particularly -k/!! and -s/!s retain state across queries,
    so a single instance of this object should be created per session, with
    handle_query() being called for each individual query.
    """
    lookup_field_names = lookup_field_names()
    database_handler: DatabaseHandler
    _current_set_root_object_class: Optional[str]

    def __init__(self, client_ip: str, client_str: str, preloader: Preloader,
                 database_handler: DatabaseHandler) -> None:
        self.all_valid_sources = list(get_setting('sources', {}).keys())
        self.sources_default = get_setting('sources_default')
        self.sources: List[str] = self.sources_default if self.sources_default else self.all_valid_sources
        if get_setting('rpki.roa_source'):
            self.all_valid_sources.append(RPKI_IRR_PSEUDO_SOURCE)
        self.object_classes: List[str] = []
        self.user_agent: Optional[str] = None
        self.multiple_command_mode = False
        self.rpki_aware = bool(get_setting('rpki.roa_source'))
        self.rpki_invalid_filter_enabled = self.rpki_aware
        self.out_scope_filter_enabled = True
        self.timeout = 30
        self.key_fields_only = False
        self.client_ip = client_ip
        self.client_str = client_str
        self.preloader = preloader
        self.database_handler = database_handler

    def handle_query(self, query: str) -> WhoisQueryResponse:
        """
        Process a single query. Always returns a WhoisQueryResponse object.
        Not thread safe - only one call must be made to this method at the same time.
        """
        # These flags are reset with every query.
        self.key_fields_only = False
        self.object_classes = []

        if query.startswith('!'):
            try:
                return self.handle_irrd_command(query[1:])
            except WhoisQueryParserException as exc:
                logger.info(f'{self.client_str}: encountered parsing error while parsing query "{query}": {exc}')
                return WhoisQueryResponse(
                    response_type=WhoisQueryResponseType.ERROR,
                    mode=WhoisQueryResponseMode.IRRD,
                    result=str(exc)
                )
            except Exception as exc:
                self.database_handler.refresh_connection()
                # Do one retry after refreshing the DB connection
                try:
                    return self.handle_irrd_command(query[1:])
                except Exception:
                    pass
                logger.error(f'An exception occurred while processing whois query "{query}": {exc}', exc_info=exc)
                return WhoisQueryResponse(
                    response_type=WhoisQueryResponseType.ERROR,
                    mode=WhoisQueryResponseMode.IRRD,
                    result='An internal error occurred while processing this query.'
                )

        try:
            return self.handle_ripe_command(query)
        except WhoisQueryParserException as exc:
            logger.info(f'{self.client_str}: encountered parsing error while parsing query "{query}": {exc}')
            return WhoisQueryResponse(
                response_type=WhoisQueryResponseType.ERROR,
                mode=WhoisQueryResponseMode.RIPE,
                result=str(exc)
            )
        except Exception as exc:
            self.database_handler.refresh_connection()
            # Do one retry after refreshing the DB connection
            try:
                return self.handle_ripe_command(query)
            except Exception:
                pass
            logger.error(f'An exception occurred while processing whois query "{query}": {exc}', exc_info=exc)
            return WhoisQueryResponse(
                response_type=WhoisQueryResponseType.ERROR,
                mode=WhoisQueryResponseMode.RIPE,
                result='An internal error occurred while processing this query.'
            )

    def handle_irrd_command(self, full_command: str) -> WhoisQueryResponse:
        """Handle an IRRD-style query. full_command should not include the first exclamation mark. """
        if not full_command:
            raise WhoisQueryParserException('Missing IRRD command')
        command = full_command[0]
        parameter = full_command[1:]
        response_type = WhoisQueryResponseType.SUCCESS
        result = None

        # A is not tested here because it is already handled in handle_irrd_routes_for_as_set
        queries_with_parameter = list('tg6ijmnors')
        if command in queries_with_parameter and not parameter:
            raise WhoisQueryParserException(f'Missing parameter for {command} query')

        if command == '!':
            self.multiple_command_mode = True
            result = None
            response_type = WhoisQueryResponseType.NO_RESPONSE
        elif full_command.upper() == 'FNO-RPKI-FILTER':
            self.rpki_invalid_filter_enabled = False
            result = 'Filtering out RPKI invalids is disabled for !r and RIPE style ' \
                     'queries for the rest of this connection.'
        elif full_command.upper() == 'FNO-SCOPE-FILTER':
            self.out_scope_filter_enabled = False
            result = 'Filtering out out-of-scope objects is disabled for !r and RIPE style ' \
                     'queries for the rest of this connection.'
        elif command == 'v':
            result = self.handle_irrd_version()
        elif command == 't':
            self.handle_irrd_timeout_update(parameter)
        elif command == 'g':
            result = self.handle_irrd_routes_for_origin_v4(parameter)
            if not result:
                response_type = WhoisQueryResponseType.KEY_NOT_FOUND
        elif command == '6':
            result = self.handle_irrd_routes_for_origin_v6(parameter)
            if not result:
                response_type = WhoisQueryResponseType.KEY_NOT_FOUND
        elif command == 'a':
            result = self.handle_irrd_routes_for_as_set(parameter)
            if not result:
                response_type = WhoisQueryResponseType.KEY_NOT_FOUND
        elif command == 'i':
            result = self.handle_irrd_set_members(parameter)
            if not result:
                response_type = WhoisQueryResponseType.KEY_NOT_FOUND
        elif command == 'j':
            result = self.handle_irrd_database_serial_range(parameter)
        elif command == 'J':
            result = self.handle_irrd_database_status(parameter)
        elif command == 'm':
            result = self.handle_irrd_exact_key(parameter)
            if not result:
                response_type = WhoisQueryResponseType.KEY_NOT_FOUND
        elif command == 'n':
            self.handle_user_agent(parameter)
        elif command == 'o':
            result = self.handle_inverse_attr_search('mnt-by', parameter)
            if not result:
                response_type = WhoisQueryResponseType.KEY_NOT_FOUND
        elif command == 'r':
            result = self.handle_irrd_route_search(parameter)
            if not result:
                response_type = WhoisQueryResponseType.KEY_NOT_FOUND
        elif command == 's':
            result = self.handle_irrd_sources_list(parameter)
        else:
            raise WhoisQueryParserException(f'Unrecognised command: {command}')

        return WhoisQueryResponse(
            response_type=response_type,
            mode=WhoisQueryResponseMode.IRRD,
            result=result,
        )

    def handle_irrd_timeout_update(self, timeout: str) -> None:
        """!timeout query - update timeout in connection"""
        try:
            timeout_value = int(timeout)
        except ValueError:
            raise WhoisQueryParserException(f'Invalid value for timeout: {timeout}')

        if timeout_value > 0 and timeout_value <= 1000:
            self.timeout = timeout_value
        else:
            raise WhoisQueryParserException(f'Invalid value for timeout: {timeout}')

    def handle_irrd_routes_for_origin_v4(self, origin: str) -> str:
        """!g query - find all originating IPv4 prefixes from an origin, e.g. !gAS65537"""
        return self._routes_for_origin(origin, 4)

    def handle_irrd_routes_for_origin_v6(self, origin: str) -> str:
        """!6 query - find all originating IPv6 prefixes from an origin, e.g. !6as65537"""
        return self._routes_for_origin(origin, 6)

    def _routes_for_origin(self, origin: str, ip_version: Optional[int]=None) -> str:
        """
        Resolve all route(6)s prefixes for an origin, returning a space-separated list
        of all originating prefixes, not including duplicates.
        """
        try:
            origin_formatted, _ = parse_as_number(origin)
        except ValidationError as ve:
            raise WhoisQueryParserException(str(ve))

        prefixes = self.preloader.routes_for_origins([origin_formatted], self.sources, ip_version=ip_version)
        return ' '.join(prefixes)

    def handle_irrd_routes_for_as_set(self, set_name: str) -> str:
        """
        !a query - find all originating prefixes for all members of an AS-set, e.g. !a4AS-FOO or !a6AS-FOO
        """
        ip_version: Optional[int] = None
        if set_name.startswith('4'):
            set_name = set_name[1:]
            ip_version = 4
        elif set_name.startswith('6'):
            set_name = set_name[1:]
            ip_version = 6

        if not set_name:
            raise WhoisQueryParserException('Missing required set name for A query')

        self._current_set_root_object_class = 'as-set'
        members = self._recursive_set_resolve({set_name})
        prefixes = self.preloader.routes_for_origins(members, self.sources, ip_version=ip_version)
        return ' '.join(prefixes)

    def handle_irrd_set_members(self, parameter: str) -> str:
        """
        !i query - find all members of an as-set or route-set, possibly recursively.
        e.g. !iAS-FOO for non-recursive, !iAS-FOO,1 for recursive
        """
        recursive = False
        if parameter.endswith(',1'):
            recursive = True
            parameter = parameter[:-2]

        self._current_set_root_object_class = None
        if not recursive:
            members, leaf_members = self._find_set_members({parameter})
            members.update(leaf_members)
        else:
            members = self._recursive_set_resolve({parameter})
        if parameter in members:
            members.remove(parameter)

        if get_setting('compatibility.ipv4_only_route_set_members'):
            original_members = set(members)
            for member in original_members:
                try:
                    IP(member)
                except ValueError:
                    continue  # This is not a prefix, ignore.
                try:
                    IP(member, ipversion=4)
                except ValueError:
                    # This was a valid prefix, but not a valid IPv4 prefix,
                    # and should be removed.
                    members.remove(member)

        return ' '.join(sorted(members))

    def _recursive_set_resolve(self, members: Set[str], sets_seen=None) -> Set[str]:
        """
        Resolve all members of a number of sets, recursively.

        For each set in members, determines whether it has been seen already (to prevent
        infinite recursion), ignores it if already seen, and then either adds
        it directly or adds it to a set that requires further resolving.
        """
        if not sets_seen:
            sets_seen = set()

        if all([member in sets_seen for member in members]):
            return set()
        sets_seen.update(members)

        set_members = set()
        resolved_as_members = set()
        sub_members, leaf_members = self._find_set_members(members)

        for sub_member in sub_members:
            if self._current_set_root_object_class is None or self._current_set_root_object_class == 'route-set':
                try:
                    IP(sub_member.split('^')[0])
                    set_members.add(sub_member)
                    continue
                except ValueError:
                    pass
            # AS numbers are permitted in route-sets and as-sets, per RFC 2622 5.3.
            # When an AS number is encountered as part of route-set resolving,
            # the prefixes originating from that AS should be added to the response.
            try:
                as_number_formatted, _ = parse_as_number(sub_member)
                if self._current_set_root_object_class == 'route-set':
                    set_members.update(self.preloader.routes_for_origins(
                        [as_number_formatted], self.sources))
                    resolved_as_members.add(sub_member)
                else:
                    set_members.add(sub_member)
                continue
            except ValueError:
                pass

        further_resolving_required = sub_members - set_members - sets_seen - resolved_as_members
        new_members = self._recursive_set_resolve(further_resolving_required, sets_seen)
        set_members.update(new_members)

        return set_members

    def _find_set_members(self, set_names: Set[str]) -> Tuple[Set[str], Set[str]]:
        """
        Find all members of a number of route-sets or as-sets. Includes both
        direct members listed in members attribute, but also
        members included by mbrs-by-ref/member-of.

        Returns a tuple of two sets:
        - members found of the sets included in set_names, both
          references to other sets and direct AS numbers, etc.
        - leaf members that were included in set_names, i.e.
          names for which no further data could be found - for
          example references to non-existent other sets
        """
        members: Set[str] = set()
        sets_already_resolved: Set[str] = set()

        columns = ['parsed_data', 'rpsl_pk', 'source', 'object_class']
        query = self._prepare_query(column_names=columns)

        object_classes = ['as-set', 'route-set']
        # Per RFC 2622 5.3, route-sets can refer to as-sets,
        # but as-sets can only refer to other as-sets.
        if self._current_set_root_object_class == 'as-set':
            object_classes = [self._current_set_root_object_class]

        query = query.object_classes(object_classes).rpsl_pks(set_names)
        query_result = list(self.database_handler.execute_query(query))

        if not query_result:
            # No sub-members are found, and apparantly all inputs were leaf members.
            return set(), set_names

        # Track the object class of the root object set.
        # In one case self._current_set_root_object_class may already be set
        # on the first run: when the set resolving should be fixed to one
        # type of set object.
        if not self._current_set_root_object_class:
            self._current_set_root_object_class = query_result[0]['object_class']

        for result in query_result:
            rpsl_pk = result['rpsl_pk']

            # The same PK may occur in multiple sources, but we are
            # only interested in the first matching object, prioritised
            # according to the source order. This priority is part of the
            # query ORDER BY, so basically we only process an RPSL pk once.
            if rpsl_pk in sets_already_resolved:
                continue
            sets_already_resolved.add(rpsl_pk)

            object_class = result['object_class']
            object_data = result['parsed_data']
            mbrs_by_ref = object_data.get('mbrs-by-ref', None)
            for members_attr in ['members', 'mp-members']:
                if members_attr in object_data:
                    members.update(set(object_data[members_attr]))

            if not rpsl_pk or not object_class or not mbrs_by_ref:
                continue

            # If mbrs-by-ref is set, find any objects with member-of pointing to the route/as-set
            # under query, and include a maintainer listed in mbrs-by-ref, unless mbrs-by-ref
            # is set to ANY.
            query_object_class = ['route', 'route6'] if object_class == 'route-set' else ['aut-num']
            query = self._prepare_query(column_names=columns).object_classes(query_object_class)
            query = query.lookup_attrs_in(['member-of'], [rpsl_pk])

            if 'ANY' not in [m.strip().upper() for m in mbrs_by_ref]:
                query = query.lookup_attrs_in(['mnt-by'], mbrs_by_ref)

            referring_objects = self.database_handler.execute_query(query)

            for result in referring_objects:
                member_object_class = result['object_class']
                members.add(result['parsed_data'][member_object_class])

        leaf_members = set_names - sets_already_resolved
        return members, leaf_members

    def handle_irrd_database_serial_range(self, parameter: str) -> str:
        """!j query - database serial range"""
        if parameter == '-*':
            sources = self.sources_default if self.sources_default else self.all_valid_sources
        else:
            sources = [s.upper() for s in parameter.split(',')]
        invalid_sources = [s for s in sources if s not in self.all_valid_sources]
        query = DatabaseStatusQuery().sources(sources)
        query_results = self.database_handler.execute_query(query)

        result_txt = ''
        for query_result in query_results:
            source = query_result['source'].upper()
            keep_journal = 'Y' if get_setting(f'sources.{source}.keep_journal') else 'N'
            serial_newest = query_result['serial_newest_mirror']
            fields = [
                source,
                keep_journal,
                f'0-{serial_newest}' if serial_newest else '-',
            ]
            if query_result['serial_last_export']:
                fields.append(str(query_result['serial_last_export']))
            result_txt += ':'.join(fields) + '\n'

        for invalid_source in invalid_sources:
            result_txt += f'{invalid_source.upper()}:X:Database unknown\n'
        return result_txt.strip()

    def handle_irrd_database_status(self, parameter: str) -> str:
        """!J query - database status"""
        if parameter == '-*':
            sources = self.sources_default if self.sources_default else self.all_valid_sources
        else:
            sources = [s.upper() for s in parameter.split(',')]
        invalid_sources = [s for s in sources if s not in self.all_valid_sources]
        query = DatabaseStatusQuery().sources(sources)
        query_results = self.database_handler.execute_query(query)

        results: OrderedDict[str, OrderedDict[str, Any]] = OrderedDict()
        for query_result in query_results:
            source = query_result['source'].upper()
            results[source] = OrderedDict()
            results[source]['authoritative'] = get_setting(f'sources.{source}.authoritative', False)
            object_class_filter = get_setting(f'sources.{source}.object_class_filter')
            results[source]['object_class_filter'] = list(object_class_filter) if object_class_filter else None
            results[source]['rpki_rov_filter'] = bool(get_setting('rpki.roa_source') and not get_setting(f'sources.{source}.rpki_excluded'))
            results[source]['scopefilter_enabled'] = bool(get_setting('scopefilter')) and not get_setting(f'sources.{source}.scopefilter_excluded')
            results[source]['local_journal_kept'] = get_setting(f'sources.{source}.keep_journal', False)
            results[source]['serial_oldest_journal'] = query_result['serial_oldest_journal']
            results[source]['serial_newest_journal'] = query_result['serial_newest_journal']
            results[source]['serial_last_export'] = query_result['serial_last_export']
            results[source]['serial_newest_mirror'] = query_result['serial_newest_mirror']
            results[source]['last_update'] = query_result['updated'].astimezone(timezone('UTC')).isoformat()
            results[source]['synchronised_serials'] = is_serial_synchronised(self.database_handler, source)

        for invalid_source in invalid_sources:
            results[invalid_source.upper()] = OrderedDict({'error': 'Unknown source'})
        return ujson.dumps(results, indent=4)

    def handle_irrd_exact_key(self, parameter: str):
        """!m query - exact object key lookup, e.g. !maut-num,AS65537"""
        try:
            object_class, rpsl_pk = parameter.split(',', maxsplit=1)
        except ValueError:
            raise WhoisQueryParserException(f'Invalid argument for object lookup: {parameter}')
        query = self._prepare_query().object_classes([object_class]).rpsl_pk(rpsl_pk).first_only()
        return self._execute_query_flatten_output(query)

    def handle_irrd_route_search(self, parameter: str):
        """
        !r query - route search with various options:
           !r192.0.2.0/24 returns all exact matching objects
           !r192.0.2.0/24,o returns space-separated origins of all exact matching objects
           !r192.0.2.0/24,l returns all one-level less specific objects, not including exact
           !r192.0.2.0/24,L returns all less specific objects, including exact
           !r192.0.2.0/24,M returns all more specific objects, not including exact
        """
        option: Optional[str] = None
        if ',' in parameter:
            address, option = parameter.split(',')
        else:
            address = parameter
        try:
            address = IP(address)
        except ValueError:
            raise WhoisQueryParserException(f'Invalid input for route search: {parameter}')

        query = self._prepare_query(ordered_by_sources=False).object_classes(['route', 'route6'])
        if option is None or option == 'o':
            query = query.ip_exact(address)
        elif option == 'l':
            query = query.ip_less_specific_one_level(address)
        elif option == 'L':
            query = query.ip_less_specific(address)
        elif option == 'M':
            query = query.ip_more_specific(address)
        else:
            raise WhoisQueryParserException(f'Invalid route search option: {option}')

        if option == 'o':
            query_result = self.database_handler.execute_query(query)
            prefixes = [r['parsed_data']['origin'] for r in query_result]
            return ' '.join(prefixes)
        return self._execute_query_flatten_output(query)

    def handle_irrd_sources_list(self, parameter: str) -> Optional[str]:
        """
        !s query - set used sources
           !s-lc returns all enabled sources, space separated
           !sripe,nttcom limits sources to ripe and nttcom
        """
        if parameter == '-lc':
            return ','.join(self.sources)

        sources = parameter.upper().split(',')
        if not all([source in self.all_valid_sources for source in sources]):
            raise WhoisQueryParserException('One or more selected sources are unavailable.')
        self.sources = sources

        return None

    def handle_irrd_version(self):
        """!v query - return version"""
        return f'IRRd -- version {__version__}'

    def handle_ripe_command(self, full_query: str) -> WhoisQueryResponse:
        """
        Process RIPE-style queries. Any query that is not explicitly an IRRD-style
        query (i.e. starts with exclamation mark) is presumed to be a RIPE query.
        """
        full_query = re.sub(' +', ' ', full_query)
        components = full_query.strip().split(' ')
        result = None
        response_type = WhoisQueryResponseType.SUCCESS
        remove_auth_hashes = True

        while len(components):
            component = components.pop(0)
            if component.startswith('-'):
                command = component[1:]
                try:
                    if command == 'k':
                        self.multiple_command_mode = True
                    elif command in ['l', 'L', 'M', 'x']:
                        result = self.handle_ripe_route_search(command, components.pop(0))
                        if not result:
                            response_type = WhoisQueryResponseType.KEY_NOT_FOUND
                        break
                    elif command == 'i':
                        result = self.handle_inverse_attr_search(components.pop(0), components.pop(0))
                        if not result:
                            response_type = WhoisQueryResponseType.KEY_NOT_FOUND
                        break
                    elif command == 's':
                        self.handle_ripe_sources_list(components.pop(0))
                    elif command == 'a':
                        self.handle_ripe_sources_list(None)
                    elif command == 'T':
                        self.handle_ripe_restrict_object_class(components.pop(0))
                    elif command == 't':
                        result = self.handle_ripe_request_object_template(components.pop(0))
                        break
                    elif command == 'K':
                        self.handle_ripe_key_fields_only()
                    elif command == 'V':
                        self.handle_user_agent(components.pop(0))
                    elif command == 'g':
                        result = self.handle_nrtm_request(components.pop(0))
                        remove_auth_hashes = False
                    elif command in ['F', 'r']:
                        continue  # These flags disable recursion, but IRRd never performs recursion anyways
                    else:
                        raise WhoisQueryParserException(f'Unrecognised flag/search: {command}')
                except IndexError:
                    raise WhoisQueryParserException(f'Missing argument for flag/search: {command}')
            else:  # assume query to be a free text search
                result = self.handle_ripe_text_search(component)

        return WhoisQueryResponse(
            response_type=response_type,
            mode=WhoisQueryResponseMode.RIPE,
            result=result,
            remove_auth_hashes=remove_auth_hashes,
        )

    def handle_ripe_route_search(self, command: str, parameter: str) -> str:
        """
        -l/L/M/x query - route search for:
           -x 192.0.2.0/2 returns all exact matching objects
           -l 192.0.2.0/2 returns all one-level less specific objects, not including exact
           -L 192.0.2.0/2 returns all less specific objects, including exact
           -M 192.0.2.0/2 returns all more specific objects, not including exact
        """
        try:
            address = IP(parameter)
        except ValueError:
            raise WhoisQueryParserException(f'Invalid input for route search: {parameter}')

        query = self._prepare_query(ordered_by_sources=False).object_classes(['route', 'route6'])
        if command == 'x':
            query = query.ip_exact(address)
        elif command == 'l':
            query = query.ip_less_specific_one_level(address)
        elif command == 'L':
            query = query.ip_less_specific(address)
        elif command == 'M':
            query = query.ip_more_specific(address)

        return self._execute_query_flatten_output(query)

    def handle_ripe_sources_list(self, sources_list: Optional[str]) -> None:
        """-s/-a parameter - set sources list. Empty list enables all sources. """
        if sources_list:
            sources = sources_list.upper().split(',')
            if not all([source in self.all_valid_sources for source in sources]):
                raise WhoisQueryParserException('One or more selected sources are unavailable.')
            self.sources = sources
        else:
            self.sources = self.sources_default if self.sources_default else self.all_valid_sources

    def handle_ripe_restrict_object_class(self, object_classes) -> None:
        """-T parameter - restrict object classes for this query, comma-seperated"""
        self.object_classes = object_classes.split(',')

    def handle_ripe_request_object_template(self, object_class) -> str:
        """-t query - return the RPSL template for an object class"""
        try:
            return OBJECT_CLASS_MAPPING[object_class]().generate_template()
        except KeyError:
            raise WhoisQueryParserException(f'Unknown object class: {object_class}')

    def handle_ripe_key_fields_only(self) -> None:
        """-K paramater - only return primary key and members fields"""
        self.key_fields_only = True

    def handle_ripe_text_search(self, value: str) -> str:
        query = self._prepare_query(ordered_by_sources=False).text_search(value)
        return self._execute_query_flatten_output(query)

    def handle_user_agent(self, user_agent: str):
        """-V/!n parameter/query - set a user agent for the client"""
        self.user_agent = user_agent
        logger.info(f'{self.client_str}: user agent set to: {user_agent}')

    def handle_nrtm_request(self, param):
        try:
            source, version, serial_range = param.split(':')
        except ValueError:
            raise WhoisQueryParserException('Invalid parameter: must contain three elements')

        try:
            serial_start, serial_end = serial_range.split('-')
            serial_start = int(serial_start)
            if serial_end == 'LAST':
                serial_end = None
            else:
                serial_end = int(serial_end)
        except ValueError:
            raise WhoisQueryParserException(f'Invalid serial range: {serial_range}')

        if version not in ['1', '3']:
            raise WhoisQueryParserException(f'Invalid NRTM version: {version}')

        source = source.upper()
        if source not in self.all_valid_sources:
            raise WhoisQueryParserException(f'Unknown source: {source}')

        in_access_list = is_client_permitted(self.client_ip, f'sources.{source}.nrtm_access_list', log=False)
        in_unfiltered_access_list = is_client_permitted(self.client_ip, f'sources.{source}.nrtm_access_list_unfiltered', log=False)
        if not in_access_list and not in_unfiltered_access_list:
            raise WhoisQueryParserException('Access denied')

        try:
            return NRTMGenerator().generate(
                source, version, serial_start, serial_end, self.database_handler,
                remove_auth_hashes=not in_unfiltered_access_list)
        except NRTMGeneratorException as nge:
            raise WhoisQueryParserException(str(nge))

    def handle_inverse_attr_search(self, attribute: str, value: str) -> str:
        """
        -i/!o query - inverse search for attribute values
        e.g. `-i mnt-by FOO` finds all objects where (one of the) maintainer(s) is FOO,
        as does `!oFOO`. Restricted to designated lookup fields.
        """
        if attribute not in self.lookup_field_names:
            readable_lookup_field_names = ', '.join(self.lookup_field_names)
            msg = (f'Inverse attribute search not supported for {attribute},' +
                   f'only supported for attributes: {readable_lookup_field_names}')
            raise WhoisQueryParserException(msg)
        query = self._prepare_query(ordered_by_sources=False).lookup_attr(attribute, value)
        return self._execute_query_flatten_output(query)

    def _prepare_query(self, column_names=None, ordered_by_sources=True) -> RPSLDatabaseQuery:
        """Prepare an RPSLDatabaseQuery by applying relevant sources/class filters."""
        query = RPSLDatabaseQuery(column_names, ordered_by_sources)
        if self.sources and self.sources != self.all_valid_sources:
            query.sources(self.sources)
        else:
            default = list(get_setting('sources_default', []))
            if default:
                query.sources(list(default))
        if self.object_classes:
            query.object_classes(self.object_classes)
        if self.rpki_invalid_filter_enabled:
            query.rpki_status([RPKIStatus.not_found, RPKIStatus.valid])
        if self.out_scope_filter_enabled:
            query.scopefilter_status([ScopeFilterStatus.in_scope])
        return query

    def _execute_query_flatten_output(self, query: RPSLDatabaseQuery) -> str:
        """
        Execute an RPSLDatabaseQuery, and flatten the output into a string with object text
        for easy passing to a WhoisQueryResponse.
        """
        query_response = self.database_handler.execute_query(query)
        if self.key_fields_only:
            result = self._filter_key_fields(query_response)
        else:
            result = ''
            for obj in query_response:
                result += obj['object_text']
                if (
                        self.rpki_aware and
                        obj['source'] != RPKI_IRR_PSEUDO_SOURCE and
                        obj['object_class'] in RPKI_RELEVANT_OBJECT_CLASSES
                ):
                    comment = ''
                    if obj['rpki_status'] == RPKIStatus.not_found:
                        comment = ' # No ROAs found, or RPKI validation not enabled for source'
                    result += f'rpki-ov-state:  {obj["rpki_status"].name}{comment}\n'
                result += '\n'
        return result.strip('\n\r')

    def _filter_key_fields(self, query_response) -> str:
        results: OrderedSet[str] = OrderedSet()
        for obj in query_response:
            result = ''
            rpsl_object_class = OBJECT_CLASS_MAPPING[obj['object_class']]
            fields_included = rpsl_object_class.pk_fields + ['members', 'mp-members']

            for field_name in fields_included:
                field_data = obj['parsed_data'].get(field_name)
                if field_data:
                    if isinstance(field_data, list):
                        for item in field_data:
                            result += f'{field_name}: {item}\n'
                    else:
                        result += f'{field_name}: {field_data}\n'
            results.add(result)
        return '\n'.join(results)



"""Set prefix_length in existing RPSL objects

Revision ID: a8609af97aa3
Revises: 64a3d6faf6d4
Create Date: 2019-03-04 16:14:17.862510

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.ext.declarative import declarative_base


# revision identifiers, used by Alembic.
revision = 'a8609af97aa3'
down_revision = '64a3d6faf6d4'
branch_labels = None
depends_on = None
Base = declarative_base()


class RPSLDatabaseObject(Base):  # type:ignore
    __tablename__ = 'rpsl_objects'

    pk = sa.Column(pg.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'), primary_key=True)
    object_class = sa.Column(sa.String, nullable=False, index=True)
    ip_size = sa.Column(sa.DECIMAL(scale=0))
    prefix_length = sa.Column(sa.Integer, nullable=True)


def upgrade():
    connection = op.get_bind()
    t_rpsl_objects = RPSLDatabaseObject.__table__

    for length in range(33):
        ip_size = pow(2, 32-length)
        connection.execute(t_rpsl_objects.update().where(
            sa.and_(t_rpsl_objects.c.ip_size == ip_size, t_rpsl_objects.c.object_class == 'route')
        ).values(
            prefix_length=length,
        ))

    for length in range(129):
        ip_size = pow(2, 128-length)
        connection.execute(t_rpsl_objects.update().where(
            sa.and_(t_rpsl_objects.c.ip_size == ip_size, t_rpsl_objects.c.object_class == 'route6')
        ).values(
            prefix_length=length,
        ))


def downgrade():
    pass



#!/usr/bin/env python
# flake8: noqa: E402
import argparse
import logging
import sys
from pathlib import Path


"""
Set last-modified attribute on all authoritative objects.
"""

logger = logging.getLogger(__name__)
sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.storage.database_handler import DatabaseHandler
from irrd.conf import config_init, CONFIG_PATH_DEFAULT, get_setting
from irrd.rpsl.rpsl_objects import rpsl_object_from_text
from irrd.storage.models import RPSLDatabaseObject
from irrd.storage.queries import RPSLDatabaseQuery

def set_last_modified():
    dh = DatabaseHandler()
    auth_sources = [k for k, v in get_setting('sources').items() if v.get('authoritative')]
    q = RPSLDatabaseQuery(column_names=['pk', 'object_text', 'updated'], enable_ordering=False)
    q = q.sources(auth_sources)

    results = list(dh.execute_query(q))
    print(f'Updating {len(results)} objects in sources {auth_sources}')
    for result in results:
        rpsl_obj = rpsl_object_from_text(result['object_text'], strict_validation=False)
        if rpsl_obj.messages.errors():  # pragma: no cover
            print(f'Failed to process {rpsl_obj}: {rpsl_obj.messages.errors()}')
            continue
        new_text = rpsl_obj.render_rpsl_text(result['updated'])
        stmt = RPSLDatabaseObject.__table__.update().where(
            RPSLDatabaseObject.__table__.c.pk == result['pk']).values(
            object_text=new_text,
        )
        dh.execute_statement(stmt)
    dh.commit()
    dh.close()


def main():  # pragma: no cover
    description = """Set last-modified attribute on all authoritative objects."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    args = parser.parse_args()

    config_init(args.config_file_path)
    if get_setting('database_readonly'):
        print('Unable to run, because database_readonly is set')
        sys.exit(-1)

    sys.exit(set_last_modified())


if __name__ == '__main__':  # pragma: no cover
    main()



import os
import sqlalchemy as sa
import ujson

from irrd.conf import get_setting

engine = None


def get_engine():
    global engine
    if engine:
        return engine
    engine = sa.create_engine(
        get_setting('database_url'),
        pool_size=2,
        json_deserializer=ujson.loads,
    )

    # https://docs.sqlalchemy.org/en/13/core/pooling.html#using-connection-pools-with-multiprocessing
    @sa.event.listens_for(engine, "connect")
    def connect(dbapi_connection, connection_record):
        connection_record.info['pid'] = os.getpid()

    @sa.event.listens_for(engine, "checkout")
    def checkout(dbapi_connection, connection_record, connection_proxy):
        pid = os.getpid()
        if connection_record.info['pid'] != pid:  # pragma: no cover
            connection_record.connection = connection_proxy.connection = None
            raise sa.exc.DisconnectionError(
                "Connection record belongs to pid %s, "
                "attempting to check out in pid %s" %
                (connection_record.info['pid'], pid)
            )

    return engine



import logging
from IPy import IP

from irrd.conf import get_setting

logger = logging.getLogger(__name__)


def is_client_permitted(ip: str, access_list_setting: str, default_deny=True, log=True) -> bool:
    """
    Determine whether a client is permitted to access an interface,
    based on the value of the setting of access_list_setting.
    If default_deny is True, an unset or empty access list will lead to denial.

    IPv6-mapped IPv4 addresses are unmapped to regular IPv4 addresses before processing.
    """
    try:
        client_ip = IP(ip)
    except (ValueError, AttributeError) as e:
        if log:
            logger.error(f'Rejecting request as client IP could not be read from '
                         f'{ip}: {e}')
        return False

    if client_ip.version() == 6:
        try:
            client_ip = client_ip.v46map()
        except ValueError:
            pass

    access_list_name = get_setting(access_list_setting)
    access_list = get_setting(f'access_lists.{access_list_name}')

    if not access_list_name or not access_list:
        if default_deny:
            if log:
                logger.info(f'Rejecting request, access list empty or undefined: {client_ip}')
            return False
        else:
            return True

    allowed = any([client_ip in IP(allowed) for allowed in access_list])
    if not allowed and log:
        logger.info(f'Rejecting request, IP not in access list {access_list_name}: {client_ip}')
    return allowed



from typing import Optional

from irrd.conf import get_setting
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.queries import RPSLDatabaseJournalQuery, DatabaseStatusQuery
from irrd.utils.text import remove_auth_hashes as remove_auth_hashes_func


class NRTMGeneratorException(Exception):
    pass


class NRTMGenerator:
    def generate(self, source: str, version: str,
                 serial_start_requested: int, serial_end_requested: Optional[int],
                 database_handler: DatabaseHandler, remove_auth_hashes=True) -> str:
        """
        Generate an NRTM response for a particular source, serial range and
        NRTM version. Raises NRTMGeneratorException for various error conditions.

        For queries where the user requested NRTM updates up to LAST,
        serial_end_requested is None.
        """
        if not get_setting(f'sources.{source}.keep_journal'):
            raise NRTMGeneratorException('No journal kept for this source, unable to serve NRTM queries')

        q = DatabaseStatusQuery().source(source)
        try:
            status = next(database_handler.execute_query(q))
        except StopIteration:
            raise NRTMGeneratorException('There are no journal entries for this source.')

        if serial_end_requested and serial_end_requested < serial_start_requested:
            raise NRTMGeneratorException(f'Start of the serial range ({serial_start_requested}) must be lower or '
                                         f'equal to end of the serial range ({serial_end_requested})')

        serial_start_available = status['serial_oldest_journal']
        serial_end_available = status['serial_newest_journal']

        if serial_start_available is None or serial_end_available is None:
            return '% Warning: there are no updates available'

        if serial_start_requested < serial_start_available:
            raise NRTMGeneratorException(f'Serials {serial_start_requested} - {serial_start_available} do not exist')

        if serial_end_requested is not None and serial_end_requested > serial_end_available:
            raise NRTMGeneratorException(f'Serials {serial_end_available} - {serial_end_requested} do not exist')

        if serial_end_requested is None:
            if serial_start_requested == serial_end_available + 1:
                # A specific message is triggered when starting from a serial
                # that is the current plus one, until LAST
                return '% Warning: there are no newer updates available'
            elif serial_start_requested > serial_end_available:
                raise NRTMGeneratorException(
                    f'Serials {serial_end_available} - {serial_start_requested} do not exist')

        serial_end_display = serial_end_available if serial_end_requested is None else serial_end_requested

        q = RPSLDatabaseJournalQuery().sources([source]).serial_range(serial_start_requested, serial_end_requested)
        operations = list(database_handler.execute_query(q))

        output = f'%START Version: {version} {source} {serial_start_requested}-{serial_end_display}\n'

        for operation in operations:
            output += '\n' + operation['operation'].value
            if version == '3':
                output += ' ' + str(operation['serial_nrtm'])
            text = operation['object_text']
            if remove_auth_hashes:
                text = remove_auth_hashes_func(text)
            output += '\n\n' + text

        output += f'\n%END {source}'
        return output



"""Add journal entry origin

Revision ID: 181670a62643
Revises: a8609af97aa3
Create Date: 2020-04-15 10:19:09.146000

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '181670a62643'
down_revision = 'a8609af97aa3'
branch_labels = None
depends_on = None


def upgrade():
    rpki_status = sa.Enum('unknown', 'mirror', 'synthetic_nrtm', 'pseudo_irr', 'auth_change', 'rpki_status', 'bogon_status', name='journalentryorigin')
    rpki_status.create(op.get_bind())

    op.add_column('rpsl_database_journal', sa.Column('origin', sa.Enum('unknown', 'mirror', 'synthetic_nrtm', 'auth_change', 'pseudo_irr', 'rpki_status', 'bogon_status', name='journalentryorigin'), server_default='unknown', nullable=False))
    op.create_index(op.f('ix_rpsl_database_journal_origin'), 'rpsl_database_journal', ['origin'], unique=False)


def downgrade():
    op.drop_index(op.f('ix_rpsl_database_journal_origin'), table_name='rpsl_database_journal')
    op.drop_column('rpsl_database_journal', 'origin')

    rpki_status = sa.Enum('unknown', 'mirror', 'synthetic_nrtm', 'auth_change', 'pseudo_irr', 'rpki_status', 'bogon_status', name='journalentryorigin')
    rpki_status.drop(op.get_bind())



import logging
from typing import Optional, List

from irrd.rpki.validators import SingleRouteROAValidator
from irrd.rpsl.parser import UnknownRPSLObjectClassException
from irrd.rpsl.rpsl_objects import rpsl_object_from_text, RPSLKeyCert
from irrd.scopefilter.validators import ScopeFilterValidator
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.models import DatabaseOperation, JournalEntryOrigin

logger = logging.getLogger(__name__)


class NRTMOperation:
    """
    NRTMOperation represents a single NRTM operation, i.e. an ADD/DEL
    with a serial number and source, from an NRTM stream.

    Note that the operation may contain an incomplete object, without a
    source attribute, but with other PK attribute(s) present.
    For deletion operations, this is permitted.
    """
    def __init__(self, source: str, operation: DatabaseOperation, serial: int, object_text: str,
                 strict_validation_key_cert: bool, rpki_aware: bool = False,
                 object_class_filter: Optional[List[str]] = None) -> None:
        self.source = source
        self.operation = operation
        self.serial = serial
        self.object_text = object_text
        self.strict_validation_key_cert = strict_validation_key_cert
        self.rpki_aware = rpki_aware
        self.object_class_filter = object_class_filter

    def save(self, database_handler: DatabaseHandler) -> bool:
        default_source = self.source if self.operation == DatabaseOperation.delete else None
        try:
            object_text = self.object_text.strip()
            # If an object turns out to be a key-cert, and strict_import_keycert_objects
            # is set, parse it again with strict validation to load it in the GPG keychain.
            obj = rpsl_object_from_text(object_text, strict_validation=False, default_source=default_source)
            if self.strict_validation_key_cert and obj.__class__ == RPSLKeyCert:
                obj = rpsl_object_from_text(object_text, strict_validation=True, default_source=default_source)

        except UnknownRPSLObjectClassException as exc:
            # Unknown object classes are only logged if they have not been filtered out.
            if not self.object_class_filter or exc.rpsl_object_class.lower() in self.object_class_filter:
                logger.info(f'Ignoring NRTM operation {str(self)}: {exc}')
            return False

        if self.object_class_filter and obj.rpsl_object_class.lower() not in self.object_class_filter:
            return False

        if obj.messages.errors():
            errors = '; '.join(obj.messages.errors())
            logger.critical(f'Parsing errors occurred while processing NRTM operation {str(self)}. '
                            f'This operation is ignored, causing potential data inconsistencies. '
                            f'A new operation for this update, without errors, '
                            f'will still be processed and cause the inconsistency to be resolved. '
                            f'Parser error messages: {errors}; original object text follows:\n{self.object_text}')
            database_handler.record_mirror_error(self.source, f'Parsing errors: {obj.messages.errors()}, '
                                                              f'original object text follows:\n{self.object_text}')
            return False

        if 'source' in obj.parsed_data and obj.parsed_data['source'].upper() != self.source:
            msg = (f'Incorrect source in NRTM object: stream has source {self.source}, found object with '
                   f'source {obj.source()} in operation {self.serial}/{self.operation.value}/{obj.pk()}. '
                   f'This operation is ignored, causing potential data inconsistencies.')
            database_handler.record_mirror_error(self.source, msg)
            logger.critical(msg)
            return False

        if self.operation == DatabaseOperation.add_or_update:
            if self.rpki_aware and obj.rpki_relevant and obj.prefix and obj.asn_first:
                roa_validator = SingleRouteROAValidator(database_handler)
                obj.rpki_status = roa_validator.validate_route(obj.prefix, obj.asn_first, obj.source())
            scope_validator = ScopeFilterValidator()
            obj.scopefilter_status, _ = scope_validator.validate_rpsl_object(obj)
            database_handler.upsert_rpsl_object(obj, JournalEntryOrigin.mirror,
                                                source_serial=self.serial)
        elif self.operation == DatabaseOperation.delete:
            database_handler.delete_rpsl_object(rpsl_object=obj, origin=JournalEntryOrigin.mirror,
                                                source_serial=self.serial)

        log = f'Completed NRTM operation {str(self)}/{obj.rpsl_object_class}/{obj.pk()}'
        if self.rpki_aware and obj.rpki_relevant:
            log += f', RPKI status {obj.rpki_status.value}'
        logger.info(log)
        return True

    def __repr__(self):
        return f'{self.source}/{self.serial}/{self.operation.value}'



import itertools
import logging
from collections import defaultdict
from typing import Dict, List, Set

from irrd.conf import get_setting
from irrd.rpki.status import RPKIStatus
from irrd.rpsl.parser import RPSLObject
from irrd.rpsl.rpsl_objects import rpsl_object_from_text
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.queries import RPSLDatabaseQuery
from irrd.utils.email import send_email

logger = logging.getLogger(__name__)


def notify_rpki_invalid_owners(database_handler: DatabaseHandler,
                               rpsl_dicts_now_invalid: List[Dict[str, str]]) -> int:
    """
    Notify the owners/contacts of newly RPKI invalid objects.

    Expects a list of objects, each a dict with their properties.
    Contacts are resolved as any mnt-nfy, or any email address on any
    tech-c or admin-c, of any maintainer of the object.
    One email is sent per email address.
    """
    if not get_setting('rpki.notify_invalid_enabled'):
        return 0

    rpsl_objs = []
    for obj in rpsl_dicts_now_invalid:
        source = obj['source']
        authoritative = get_setting(f'sources.{source}.authoritative')
        if authoritative and obj['rpki_status'] == RPKIStatus.invalid:
            rpsl_objs.append(rpsl_object_from_text(obj['object_text']))

    if not rpsl_objs:
        return 0

    sources = set([obj.parsed_data['source'] for obj in rpsl_objs])
    mntner_emails_by_source = {}
    for source in sources:
        # For each source, a multi-step process is run to fill this
        # dict with the contact emails for each mntner.
        mntner_emails = defaultdict(set)

        # Step 1: retrieve all relevant maintainers from the DB
        mntner_pks = set(itertools.chain(*[
            obj.parsed_data.get('mnt-by', [])
            for obj in rpsl_objs
            if obj.parsed_data['source'] == source
        ]))
        query = RPSLDatabaseQuery(['rpsl_pk', 'parsed_data']).sources([source]).rpsl_pks(mntner_pks).object_classes(['mntner'])
        mntners = list(database_handler.execute_query(query))

        # Step 2: any mnt-nfy on these maintainers is a contact address
        for mntner in mntners:
            mntner_emails[mntner['rpsl_pk']].update(mntner['parsed_data'].get('mnt-nfy', []))

        # Step 3: extract the contact handles for each maintainer
        mntner_contacts = {
            m['rpsl_pk']: m['parsed_data'].get('tech-c', []) + m['parsed_data'].get('admin-c', [])
            for m in mntners
        }

        # Step 4: retrieve all these contacts from the DB in bulk,
        # and extract their e-mail addresses
        contact_pks = set(itertools.chain(*mntner_contacts.values()))
        query = RPSLDatabaseQuery(['rpsl_pk', 'parsed_data']).sources([source]).rpsl_pks(contact_pks).object_classes(['role', 'person'])
        contacts = {
            r['rpsl_pk']: r['parsed_data'].get('e-mail', [])
            for r in database_handler.execute_query(query)
        }

        # Step 5: use the contacts per maintainer, and emails per contact
        # to create a flattened list of emails per maintainer
        for mntner_pk, mntner_contacts in mntner_contacts.items():
            for contact_pk in mntner_contacts:
                try:
                    mntner_emails[mntner_pk].update(contacts[contact_pk])
                except KeyError:
                    pass

        mntner_emails_by_source[source] = mntner_emails

    # With mntners_emails_by_source filled with per source, per maintainer,
    # all relevant emails, categorise the RPSL objects on which email
    # addresses they need to be sent to.
    objs_per_email: Dict[str, Set[RPSLObject]] = defaultdict(set)
    for rpsl_obj in rpsl_objs:
        mntners = rpsl_obj.parsed_data.get('mnt-by', [])
        source = rpsl_obj.parsed_data['source']
        for mntner_pk in mntners:
            try:
                for email in mntner_emails_by_source[source][mntner_pk]:
                    objs_per_email[email].add(rpsl_obj)
            except KeyError:  # pragma: no cover
                pass

    header_template = get_setting('rpki.notify_invalid_header', '')
    subject_template = get_setting('rpki.notify_invalid_subject', '').replace('\n', ' ')
    for email, objs in objs_per_email.items():
        sources_str = ', '.join(set([obj.parsed_data['source'] for obj in objs]))
        subject = subject_template.format(sources_str=sources_str, object_count=len(objs))
        body = header_template.format(sources_str=sources_str, object_count=len(objs))
        body += '\nThe following objects are affected:\n'
        body += '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n'
        for rpsl_obj in objs:
            body += rpsl_obj.render_rpsl_text() + '\n'
        body += '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
        try:
            send_email(email, subject, body)
        except Exception as e:  # pragma: no cover
            logger.warning(f'Unable to send RPKI invalid notification to {email}: {e}')

    return len(objs_per_email.keys())



import logging
import random
import signal
import threading
import time
from collections import defaultdict
from typing import Optional, List, Set, Dict, Union

import redis
from setproctitle import setproctitle

from irrd.conf import get_setting
from irrd.rpki.status import RPKIStatus
from irrd.scopefilter.status import ScopeFilterStatus
from irrd.utils.process_support import ExceptionLoggingProcess
from .queries import RPSLDatabaseQuery

SENTINEL_HASH_CREATED = b'SENTINEL_HASH_CREATED'
REDIS_ORIGIN_ROUTE4_STORE_KEY = b'irrd-preload-origin-route4'
REDIS_ORIGIN_ROUTE6_STORE_KEY = b'irrd-preload-origin-route6'
REDIS_PRELOAD_RELOAD_CHANNEL = 'irrd-preload-reload-channel'
REDIS_PRELOAD_COMPLETE_CHANNEL = 'irrd-preload-complete-channel'
REDIS_ORIGIN_LIST_SEPARATOR = ','
REDIS_KEY_ORIGIN_SOURCE_SEPARATOR = '_'
MAX_MEMORY_LIFETIME = 60

logger = logging.getLogger(__name__)

"""
The preloader allows information to be preloaded into memory.

For queries that repeat often, or repeatedly retrieve nearly the same,
large data sets, this can improve performance.
"""


class PersistentPubSubWorkerThread(redis.client.PubSubWorkerThread):  # type: ignore
    def __init__(self, callback, *args, **kwargs):
        self.callback = callback
        self.should_resubscribe = True
        super().__init__(*args, **kwargs)

    def run(self):
        self._running.set()
        while self._running.is_set():
            try:
                if self.should_resubscribe:
                    self.pubsub.subscribe(**{REDIS_PRELOAD_COMPLETE_CHANNEL: self.callback})
                    self.should_resubscribe = False
                self.pubsub.get_message(ignore_subscribe_messages=True, timeout=self.sleep_time)
            except redis.ConnectionError as rce:  # pragma: no cover
                logger.error(f'Failed redis pubsub connection, '
                             f'attempting reconnect and reload in 5s: {rce}')
                time.sleep(5)
                self.should_resubscribe = True
            except Exception as exc:  # pragma: no cover
                logger.error(
                    f'Error while loading in-memory preload, attempting reconnect and reload in 5s,'
                    f'traceback follows: {exc}', exc_info=exc)
                time.sleep(5)
                self.should_resubscribe = True
        self.pubsub.close()  # pragma: no cover


class Preloader:
    """
    A preloader object provides access to the preload store.
    It can be used to query the store, or signal that the store
    needs to be updated. This interface can be used from any thread
    or process.
    """
    _memory_loaded = False

    def __init__(self, enable_queries=True):
        """
        Initialise the preloader.
        If this instance is only used for signalling that the store needs to be
        updated, set enable_queries=False.
        Otherwise, this method starts a background thread that keeps an in-memory store,
        which is automatically updated.
        """
        self._redis_conn = redis.Redis.from_url(get_setting('redis_url'))
        if enable_queries:
            self._pubsub = self._redis_conn.pubsub()
            self._pubsub_thread = PersistentPubSubWorkerThread(
                callback=self._load_routes_into_memory,
                pubsub=self._pubsub,
                sleep_time=5,
                daemon=True
            )
            self._pubsub_thread.start()
            if get_setting('database_readonly'):  # pragma: no cover
                # If this instance is readonly, another IRRd process will be updating
                # the store, and likely has already done so, meaning we can try to load
                # from Redis right away instead of waiting for a signal.
                self._load_routes_into_memory()

    def signal_reload(self, object_classes_changed: Optional[Set[str]]=None) -> None:
        """
        Perform a (re)load.
        Should be called after changes to the DB have been committed.

        This will signal the process running PreloadStoreManager to reload
        the store.

        If object_classes_changed is provided, a reload is only performed
        if those classes are relevant to the data in the preload store.
        """
        relevant_object_classes = {'route', 'route6'}
        if object_classes_changed is None or object_classes_changed.intersection(relevant_object_classes):
            self._redis_conn.publish(REDIS_PRELOAD_RELOAD_CHANNEL, 'reload')

    def routes_for_origins(self, origins: Union[List[str], Set[str]], sources: List[str],
                           ip_version: Optional[int] = None) -> Set[str]:
        """
        Retrieve all prefixes (in str format) originating from the provided origins,
        from the given sources.

        Prefixes are guaranteed to be unique. ip_version can be set to 4 or 6
        to restrict responses to IPv4 or IPv6 prefixes. Blocks until the first
        store has been built.
        Origins must be strings in a cleaned format, e.g. AS65537, but not
        AS065537 or as65537.
        This call will block until the preload store is loaded.
        """
        while not self._memory_loaded:
            time.sleep(1)  # pragma: no cover
        if ip_version and ip_version not in [4, 6]:
            raise ValueError(f'Invalid IP version: {ip_version}')
        if not origins or not sources:
            return set()

        prefix_sets: Set[str] = set()
        for source in sources:
            for origin in origins:
                if (not ip_version or ip_version == 4) and source in self._origin_route4_store and origin in self._origin_route4_store[source]:
                    prefix_sets.update(self._origin_route4_store[source][origin].split(REDIS_ORIGIN_LIST_SEPARATOR))
                if (not ip_version or ip_version == 6) and source in self._origin_route6_store and origin in self._origin_route6_store[source]:
                    prefix_sets.update(self._origin_route6_store[source][origin].split(REDIS_ORIGIN_LIST_SEPARATOR))

        return prefix_sets

    def _load_routes_into_memory(self, redis_message=None):
        """
        Update the in-memory store. This is called whenever a
        message is sent to REDIS_PRELOAD_COMPLETE_CHANNEL.
        """
        while not self._redis_conn.exists(REDIS_ORIGIN_ROUTE4_STORE_KEY):
            time.sleep(1)  # pragma: no cover

        # Create a bit of randomness in when workers will update
        time.sleep(random.random())

        new_origin_route4_store = dict()
        new_origin_route6_store = dict()

        def _load(redis_key, target):
            for key, routes in self._redis_conn.hgetall(redis_key).items():
                if key == SENTINEL_HASH_CREATED:
                    continue
                source, origin = key.decode('ascii').split(REDIS_KEY_ORIGIN_SOURCE_SEPARATOR)
                if source not in target:
                    target[source] = dict()
                target[source][origin] = routes.decode('ascii')

        _load(REDIS_ORIGIN_ROUTE4_STORE_KEY, new_origin_route4_store)
        _load(REDIS_ORIGIN_ROUTE6_STORE_KEY, new_origin_route6_store)

        self._origin_route4_store = new_origin_route4_store
        self._origin_route6_store = new_origin_route6_store

        self._memory_loaded = True


class PreloadStoreManager(ExceptionLoggingProcess):
    """
    The preload store manager manages the preloaded data store, and ensures
    it is created and updated.
    There should only be one of these per IRRd instance.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._target = self.main
        self._redis_conn = redis.Redis.from_url(get_setting('redis_url'))

    def main(self):
        """
        Main function for the preload manager.

        Monitors a Redis pubsub channel, and triggers a reload when
        a message is received.
        """
        setproctitle('irrd-preload-store-manager')
        try:
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
        except ValueError:
            # During tests, this is run from a thread,
            # which does not allow setting signal handlers.
            pass
        logging.info('Starting preload store manager')

        self._clear_existing_data()
        self._pubsub = self._redis_conn.pubsub()

        self._reload_lock = threading.Lock()
        self._threads = []
        self.terminate = False  # Used to exit main() in tests

        while not self.terminate:
            self.perform_reload()
            try:
                self._pubsub.subscribe(REDIS_PRELOAD_RELOAD_CHANNEL)
                for item in self._pubsub.listen():
                    if item['type'] == 'message':
                        logger.debug('Reload requested through redis channel')
                        self.perform_reload()
                        if self.terminate:
                            return
            except redis.ConnectionError as rce:  # pragma: no cover
                logger.error(f'Failed redis pubsub connection, attempting reconnect and reload in 5s: {rce}')
                time.sleep(5)

    def _clear_existing_data(self) -> None:
        """
        Clear the existing data. This is done on startup, to ensure no
        queries are being answered with outdated data.
        """
        try:
            self._redis_conn.delete(REDIS_ORIGIN_ROUTE4_STORE_KEY, REDIS_ORIGIN_ROUTE6_STORE_KEY)
        except redis.ConnectionError as rce:  # pragma: no cover
            logger.error(f'Failed to empty preload store due to redis connection error, '
                         f'queries may have outdated results until full reload is completed (max 30s): {rce}')

    def perform_reload(self) -> None:
        """
        Perform a (re)load.
        Should be called after changes to the DB have been committed.

        This will start a new thread to reload the store. If a thread is
        already running, the new thread will start after the current one
        is done, due to locking.

        If a current thread is running, and a next thread is already
        running as well (waiting for a lock) no action is taken. The
        change that prompted this reload call will already be processed
        by the thread that is currently waiting.
        """
        self._remove_dead_threads()
        if len(self._threads) > 1:
            # Another thread is already scheduled to follow the current one
            return
        thread = PreloadUpdater(self, self._reload_lock)
        thread.start()
        self._threads.append(thread)

    def update_route_store(self, new_origin_route4_store, new_origin_route6_store) -> bool:
        """
        Store the new route information in redis. Returns True on success, False on failure.
        """
        try:
            pipeline = self._redis_conn.pipeline(transaction=True)
            pipeline.delete(REDIS_ORIGIN_ROUTE4_STORE_KEY, REDIS_ORIGIN_ROUTE6_STORE_KEY)
            # The redis store can't store sets, only strings
            origin_route4_str_dict = {k: REDIS_ORIGIN_LIST_SEPARATOR.join(v) for k, v in new_origin_route4_store.items()}
            origin_route6_str_dict = {k: REDIS_ORIGIN_LIST_SEPARATOR.join(v) for k, v in new_origin_route6_store.items()}
            # Redis can't handle empty dicts, but the dict needs to be present
            # in order not to block queries.
            origin_route4_str_dict[SENTINEL_HASH_CREATED] = '1'
            origin_route6_str_dict[SENTINEL_HASH_CREATED] = '1'
            # hmset causes a deprecation warning, but is required for Redis 3 compatibility
            pipeline.hmset(REDIS_ORIGIN_ROUTE4_STORE_KEY, origin_route4_str_dict)
            pipeline.hmset(REDIS_ORIGIN_ROUTE6_STORE_KEY, origin_route6_str_dict)
            pipeline.execute()

            self._redis_conn.publish(REDIS_PRELOAD_COMPLETE_CHANNEL, 'complete')
            return True

        except redis.ConnectionError as rce:  # pragma: no cover
            logger.error(f'Failed to update preload store due to redis connection error, '
                         f'attempting new reload in 5s: {rce}')
            time.sleep(5)
            self.perform_reload()
            return False

    def _remove_dead_threads(self) -> None:
        """
        Remove dead threads from self.threads(),
        for an accurate count of how many are alive.
        """
        for thread in self._threads:
            if not thread.is_alive():
                self._threads.remove(thread)


class PreloadUpdater(threading.Thread):
    """
    PreloadUpdater is a thread that updates the preload store,
    currently for prefixes per origin per address family.
    It is started by PreloadStoreManager.
    """

    def __init__(self, preloader, reload_lock, *args, **kwargs):
        self.preloader = preloader
        self.reload_lock = reload_lock
        super().__init__(*args, **kwargs)

    def run(self, mock_database_handler=None) -> None:
        """
        Main thread runner function.

        The reload_lock ensures only a single instance can run at the same
        time, i.e. if two threads are started, one will wait for the other
        to finish.

        For tests, mock_database_handler can be used to provide a mock.
        """
        self.reload_lock.acquire()
        try:
            self.update(mock_database_handler)
        except Exception as exc:
            logger.critical(f'Updating preload store failed, retrying in 5s, '
                            f'traceback follows: {exc}', exc_info=exc)
            time.sleep(5)
            self.preloader.perform_reload()
        finally:
            self.reload_lock.release()

    def update(self, mock_database_handler=None) -> None:
        """
        Update the store.

        After loading the data from the database, sets the two new stores
        on the provided preloader object.
        The lock is then released to allow another thread to start, and
        the store_ready_event set to indicate that the store has been
        loaded at least once, and answers can be provided based on it.
        """
        logger.debug(f'Starting preload store update from thread {self}')

        new_origin_route4_store: Dict[str, set] = defaultdict(set)
        new_origin_route6_store: Dict[str, set] = defaultdict(set)

        if not mock_database_handler:  # pragma: no cover
            from .database_handler import DatabaseHandler
            dh = DatabaseHandler(readonly=True)
        else:
            dh = mock_database_handler

        q = RPSLDatabaseQuery(column_names=['ip_version', 'ip_first', 'prefix_length', 'asn_first', 'source'], enable_ordering=False)
        q = q.object_classes(['route', 'route6']).rpki_status([RPKIStatus.not_found, RPKIStatus.valid])
        q = q.scopefilter_status([ScopeFilterStatus.in_scope])

        for result in dh.execute_query(q):
            prefix = result['ip_first']
            key = result['source'] + REDIS_KEY_ORIGIN_SOURCE_SEPARATOR + 'AS' + str(result['asn_first'])
            length = result['prefix_length']

            if result['ip_version'] == 4:
                new_origin_route4_store[key].add(f'{prefix}/{length}')
            if result['ip_version'] == 6:
                new_origin_route6_store[key].add(f'{prefix}/{length}')

        dh.close()

        if self.preloader.update_route_store(new_origin_route4_store, new_origin_route6_store):
            logger.info(f'Completed updating preload store from thread {self}')



#!/usr/bin/env python
# flake8: noqa: E402
import sys

import argparse
from alembic import command
from alembic.config import Config
from pathlib import Path

irrd_root = str(Path(__file__).resolve().parents[2])
sys.path.append(irrd_root)

from irrd.conf import config_init, CONFIG_PATH_DEFAULT


def run(version):
    alembic_cfg = Config()
    alembic_cfg.set_main_option('script_location', f'{irrd_root}/irrd/storage/alembic')
    command.upgrade(alembic_cfg, version)
    print(f'Upgrade successful, or already on latest version.')


def main():  # pragma: no cover
    description = """Upgrade the IRRd SQL database to a particular version by running database migrations."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    parser.add_argument('--version', dest='version', type=str, default='head',
                        help=f'version to upgrade to (default: head, i.e. latest)')
    args = parser.parse_args()

    config_init(args.config_file_path)
    run(args.version)


if __name__ == '__main__':  # pragma: no cover
    main()



from enum import Enum
from typing import Optional

from irrd.utils.text import remove_auth_hashes


class WhoisQueryResponseType(Enum):
    """
    Types of responses to queries.
    KEY_NOT_FOUND is specific to IRRD-style.
    NO_RESPONSE means no response should be sent at all.
    """
    SUCCESS = 'success'
    ERROR = 'error'
    KEY_NOT_FOUND = 'key_not_found'
    NO_RESPONSE = 'no_response'


class WhoisQueryResponseMode(Enum):
    """Response mode for queries - IRRD and RIPE queries have different output."""
    IRRD = 'irrd'
    RIPE = 'ripe'


class WhoisQueryResponse:
    """
    Container for all data for a response to a query.

    Based on the response_type and mode, can render a string of the complete
    response to send back to the user.
    """
    response_type: WhoisQueryResponseType = WhoisQueryResponseType.SUCCESS
    mode: WhoisQueryResponseMode = WhoisQueryResponseMode.RIPE
    result: Optional[str] = None

    def __init__(
            self,
            response_type: WhoisQueryResponseType,
            mode: WhoisQueryResponseMode,
            result: Optional[str],
            remove_auth_hashes=True,
    ) -> None:
        self.response_type = response_type
        self.mode = mode
        self.result = result
        self.remove_auth_hashes = remove_auth_hashes

    def generate_response(self) -> str:
        self.clean_response()

        if self.mode == WhoisQueryResponseMode.IRRD:
            response = self._generate_response_irrd()
            if response is not None:
                return response

        elif self.mode == WhoisQueryResponseMode.RIPE:
            response = self._generate_response_ripe()
            if response is not None:
                return response

        raise RuntimeError(f'Unable to formulate response for {self.response_type} / {self.mode}: {self.result}')

    def clean_response(self):
        if self.remove_auth_hashes:
            self.result = remove_auth_hashes(self.result)

    def _generate_response_irrd(self) -> Optional[str]:
        if self.response_type == WhoisQueryResponseType.SUCCESS:
            if self.result:
                result_len = len(self.result) + 1
                return f'A{result_len}\n{self.result}\nC\n'
            else:
                return 'C\n'
        elif self.response_type == WhoisQueryResponseType.KEY_NOT_FOUND:
            return 'D\n'
        elif self.response_type == WhoisQueryResponseType.ERROR:
            return f'F {self.result}\n'
        elif self.response_type == WhoisQueryResponseType.NO_RESPONSE:
            return ''
        return None

    def _generate_response_ripe(self) -> Optional[str]:
        # RIPE-style responses need two empty lines at the end, hence
        # the multiple newlines for each response (#335)
        # # https://www.ripe.net/manage-ips-and-asns/db/support/documentation/ripe-database-query-reference-manual#2-0-querying-the-ripe-database
        if self.response_type == WhoisQueryResponseType.SUCCESS:
            if self.result:
                return self.result + '\n\n\n'
            return '%  No entries found for the selected source(s).\n\n\n'
        elif self.response_type == WhoisQueryResponseType.KEY_NOT_FOUND:
            return '%  No entries found for the selected source(s).\n\n\n'
        elif self.response_type == WhoisQueryResponseType.ERROR:
            return f'%% ERROR: {self.result}\n\n\n'
        return None



# flake8: noqa: W293
import logging
import textwrap
from typing import Optional

from irrd.utils import email
from .handler import ChangeSubmissionHandler

logger = logging.getLogger(__name__)


def handle_email_submission(email_txt: str) -> Optional[ChangeSubmissionHandler]:
    handler = None
    try:
        msg = email.EmailParser(email_txt)
        request_meta = {
            'Message-ID': msg.message_id,
            'From': msg.message_from,
            'Date': msg.message_date,
            'Subject': msg.message_subject,
        }
    except Exception as exc:
        logger.critical(f'An exception occurred while attempting to parse the following update e-mail: {email_txt}\n'
                        f'--- traceback for {exc} follows:', exc_info=exc)
        return None

    if not msg.message_from:
        logger.critical(f'No from address was found while attempting to parse the following update e-mail - '
                        f'update not processed: {email_txt}\n')
        return None

    try:
        if not msg.body:
            logger.warning(f'Unable to extract message body from e-mail {msg.message_id} from {msg.message_from}')
            subject = f'FAILED: {msg.message_subject}'
            reply_content = textwrap.dedent(f"""
            Unfortunately, your message with ID {msg.message_id}
            could not be processed, as no text/plain part could be found.
            
            Please try to resend your message as plain text email.
            """)
        else:
            handler = ChangeSubmissionHandler(msg.body, msg.pgp_fingerprint, request_meta)
            logger.info(f'Processed e-mail {msg.message_id} from {msg.message_from}: {handler.status()}')
            logger.debug(f'Report for e-mail {msg.message_id} from {msg.message_from}: {handler.submitter_report()}')

            subject = f'{handler.status()}: {msg.message_subject}'
            reply_content = handler.submitter_report()

    except Exception as exc:
        logger.critical(f'An exception occurred while attempting to process the following update: {email_txt}\n'
                        f'--- traceback for {exc} follows:', exc_info=exc)
        subject = f'ERROR: {msg.message_subject}'
        reply_content = textwrap.dedent(f"""
        Unfortunately, your message with ID {msg.message_id}
        could not be processed, due to an internal error.
        """)

    try:
        email.send_email(msg.message_from, subject, reply_content)
        if handler:
            handler.send_notification_target_reports()
    except Exception as exc:
        logger.critical(f'An exception occurred while attempting to send a reply to an update: '
                        f'{subject}\n{reply_content}\n --- traceback for {exc} follows:', exc_info=exc)

    return handler





import logging
import socket

import email
from email.mime.text import MIMEText
from smtplib import SMTP
from typing import Optional

from irrd import __version__
from irrd.conf import get_setting
from irrd.utils.pgp import validate_pgp_signature


logger = logging.getLogger(__name__)


class EmailParser:
    """
    Parse a raw email.
    """
    body: Optional[str] = None
    pgp_fingerprint: Optional[str] = None
    message_id: Optional[str] = None
    message_from: Optional[str] = None
    message_date: Optional[str] = None
    message_subject: Optional[str] = None
    _default_encoding = 'ascii'
    _raw_body: Optional[str] = None
    _pgp_signature: Optional[str] = None

    def __init__(self, email_txt: str) -> None:
        """
        Extract data from an MIME email message.

        If present, the following fields will be filled:
        - body: the processed message body (decoded, only PGP signed parts included if applicable)
        - pgp_fingerprint: the hex-encoded fingerprint of the PGP key that signed body
        - message_id/message_from/message_date/message_subject: the values of these headers in the message
        """
        message: email.message.Message = email.message_from_string(email_txt)

        # self._raw_body will include headers like Content-Type, self.body contains the decoded contents
        if message.is_multipart():
            for part in message.walk():
                charset = part.get_content_charset()
                if not charset:
                    charset = self._default_encoding
                if part.get_content_type() == 'text/plain':
                    self.body = part.get_payload(decode=True).decode(str(charset), 'ignore')  # type: ignore
                    self._raw_body = part.as_string()
                if part.get_content_type() == 'application/pgp-signature':
                    self._pgp_signature = part.get_payload(decode=True).decode(str(charset), 'backslashreplace').strip()  # type: ignore
        else:
            charset = message.get_content_charset(failobj=self._default_encoding)
            self.body = message.get_payload(decode=True).decode(charset, 'backslashreplace')  # type: ignore
            self._raw_body = self.body

        self.message_id = message['Message-ID']  # type: ignore
        self.message_from = message['From']  # type: ignore
        self.message_date = message['Date']  # type: ignore
        self.message_subject = message['Subject']  # type: ignore

        if not (self.body and self._raw_body):
            return

        new_body, self.pgp_fingerprint = validate_pgp_signature(self._raw_body, self._pgp_signature)
        if new_body:
            self.body = new_body


def send_email(recipient, subject, body) -> None:
    if get_setting('email.recipient_override'):
        recipient = get_setting('email.recipient_override')

    logger.debug(f'Sending email to {recipient}, subject {subject}')
    body += get_setting('email.footer')
    hostname = socket.gethostname()
    body += f'\n\nGenerated by IRRd version {__version__} on {hostname}'

    msg = MIMEText(body)
    msg['Subject'] = subject
    msg['From'] = get_setting('email.from')
    msg['To'] = recipient

    s = SMTP(get_setting('email.smtp'))
    s.send_message(msg)
    s.quit()



#!/usr/bin/env python
# flake8: noqa: E402
import sys
import time
from pwd import getpwnam

import argparse
import daemon
import logging
import os
import signal
from pathlib import Path
from pid import PidFile, PidFileError


logger = logging.getLogger(__name__)
sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd import __version__
from irrd.conf import config_init, CONFIG_PATH_DEFAULT, get_setting, get_configuration
from irrd.mirroring.scheduler import MirrorScheduler
from irrd.server.http.server import start_http_server
from irrd.server.whois.server import start_whois_server
from irrd.storage.preload import PreloadStoreManager
from irrd.utils.process_support import ExceptionLoggingProcess
from irrd.storage.preload import PreloadStoreManager
from irrd.server.whois.server import start_whois_server
from irrd.server.http.server import start_http_server
from irrd.mirroring.scheduler import MirrorScheduler
from irrd.conf import config_init, CONFIG_PATH_DEFAULT, get_setting, get_configuration
from irrd import __version__


# This file does not have a unit test, but is instead tested through
# the integration tests. Writing a unit test would be too complex.

def main():
    description = """IRRd main process"""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    parser.add_argument('--foreground', dest='foreground', action='store_true',
                        help=f"run IRRd in the foreground, don't detach")
    args = parser.parse_args()

    mirror_frequency = int(os.environ.get('IRRD_SCHEDULER_TIMER_OVERRIDE', 15))

    daemon_kwargs = {
        'umask': 0o022,
    }
    if args.foreground:
        daemon_kwargs['detach_process'] = False
        daemon_kwargs['stdout'] = sys.stdout
        daemon_kwargs['stderr'] = sys.stderr

    # config_init w/ commit may only be called within DaemonContext
    config_init(args.config_file_path, commit=False)

    with daemon.DaemonContext(**daemon_kwargs):
        config_init(args.config_file_path)
        piddir = get_setting('piddir')
        logger.info('IRRd attempting to secure PID')
        try:
            with PidFile(pidname='irrd', piddir=piddir):
                logger.info(f'IRRd {__version__} starting, PID {os.getpid()}, PID file in {piddir}')
                run_irrd(mirror_frequency)
        except PidFileError as pfe:
            logger.error(f'Failed to start IRRd, unable to lock PID file irrd.pid in {piddir}: {pfe}')
        except Exception as e:
            logger.error(f'Error occurred in main process, terminating. Error follows:')
            logger.exception(e)
            os.kill(os.getpid(), signal.SIGTERM)


def run_irrd(mirror_frequency: int):
    terminated = False

    mirror_scheduler = MirrorScheduler()
    whois_process = ExceptionLoggingProcess(target=start_whois_server, name='irrd-whois-server-listener')
    whois_process.start()
    http_process = ExceptionLoggingProcess(target=start_http_server, name='irrd-http-server-listener')
    http_process.start()

    preload_manager = None
    if not get_setting(f'database_readonly'):
        preload_manager = PreloadStoreManager(name='irrd-preload-store-manager')
        preload_manager.start()

    def sighup_handler(signum, frame):
        # On SIGHUP, check if the configuration is valid and reload in
        # this process, and if it is valid, signal our three long-running
        # child processes. All other processes are short-lived and forked
        # from those or this process, so any new ones will pick up
        # the new config automatically.
        if get_configuration().reload():
            pids = [whois_process.pid, http_process.pid, preload_manager.pid]
            logging.info('Main process received SIGHUP with valid config, sending SIGHUP to '
                         f'child processes {pids}')
            for pid in pids:
                os.kill(pid, signal.SIGHUP)
    signal.signal(signal.SIGHUP, sighup_handler)

    def sigterm_handler(signum, frame):
        logging.info(f'Main process received SIGTERM, sending SIGTERM to all child processes')
        mirror_scheduler.terminate_children()
        try:
            whois_process.terminate()
        except Exception:  # pragma no cover
            pass
        try:
            http_process.terminate()
        except Exception:  # pragma no cover
            pass
        try:
            preload_manager.terminate()
        except Exception:  # pragma no cover
            pass
        nonlocal terminated
        terminated = True
    signal.signal(signal.SIGTERM, sigterm_handler)

    sleeps = mirror_frequency
    while not terminated:
        # This loops every second to prevent long blocking on SIGTERM.
        mirror_scheduler.update_process_state()
        if sleeps >= mirror_frequency:
            mirror_scheduler.run()
            sleeps = 0
        time.sleep(1)
        sleeps += 1

    logging.debug(f'Main process waiting for child processes to terminate')
    whois_process.join()
    http_process.join()
    if preload_manager:
        preload_manager.join()
    logging.info(f'Main process exiting')


if __name__ == '__main__':  # pragma: no cover
    main()



import time
from collections import defaultdict

import logging
import multiprocessing

import signal
from setproctitle import setproctitle
from typing import Dict

from irrd.conf import get_setting, RPKI_IRR_PSEUDO_SOURCE
from irrd.conf.defaults import DEFAULT_SOURCE_IMPORT_TIMER, DEFAULT_SOURCE_EXPORT_TIMER
from .mirror_runners_export import SourceExportRunner
from .mirror_runners_import import RPSLMirrorImportUpdateRunner, ROAImportRunner, \
    ScopeFilterUpdateRunner

logger = logging.getLogger(__name__)

MAX_SIMULTANEOUS_RUNS = 3


class ScheduledTaskProcess(multiprocessing.Process):
    def __init__(self, runner, *args, **kwargs):
        self.runner = runner
        super().__init__(*args, **kwargs)

    def close(self):  # pragma: no cover
        """
        close() is not available in Python 3.6,
        use our own implementation if needed.
        """
        if hasattr(super, 'close'):
            return super().close()
        if self._popen is not None:
            if self._popen.poll() is None:
                raise ValueError("Cannot close a process while it is still running. "
                                 "You should first call join() or terminate().")
            self._popen = None
            del self._sentinel
        self._closed = True

    def run(self):
        # Disable the special sigterm_handler defined in main()
        # (signal handlers are inherited)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        setproctitle(f'irrd-{self.name}')
        self.runner.run()


class MirrorScheduler:
    """
    Scheduler for mirroring processes.

    For each time run() is called, will start a process for each mirror database
    unless a process is still running for that database (which is likely to be
    the case in some full imports).
    """
    processes: Dict[str, ScheduledTaskProcess]
    last_started_time: Dict[str, int]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.processes = dict()
        self.last_started_time = defaultdict(lambda: 0)
        self.previous_scopefilter_prefixes = None
        self.previous_scopefilter_asns = None
        self.previous_scopefilter_excluded = None

    def run(self) -> None:
        if get_setting('database_readonly'):
            return

        if get_setting('rpki.roa_source'):
            import_timer = int(get_setting('rpki.roa_import_timer'))
            self.run_if_relevant(RPKI_IRR_PSEUDO_SOURCE, ROAImportRunner, import_timer)

        if self._check_scopefilter_change():
            self.run_if_relevant('scopefilter', ScopeFilterUpdateRunner, 0)

        sources_started = 0
        for source in get_setting('sources', {}).keys():
            if sources_started >= MAX_SIMULTANEOUS_RUNS:
                break
            started_import = False
            started_export = False

            is_mirror = get_setting(f'sources.{source}.import_source') or get_setting(f'sources.{source}.nrtm_host')
            import_timer = int(get_setting(f'sources.{source}.import_timer', DEFAULT_SOURCE_IMPORT_TIMER))

            if is_mirror:
                started_import = self.run_if_relevant(source, RPSLMirrorImportUpdateRunner, import_timer)

            runs_export = get_setting(f'sources.{source}.export_destination') or get_setting(f'sources.{source}.export_destination_unfiltered')
            export_timer = int(get_setting(f'sources.{source}.export_timer', DEFAULT_SOURCE_EXPORT_TIMER))

            if runs_export:
                started_export = self.run_if_relevant(source, SourceExportRunner, export_timer)

            if started_import or started_export:
                sources_started += 1

    def _check_scopefilter_change(self) -> bool:
        """
        Check whether the scope filter has changed since last call.
        Always returns True on the first call.
        """
        if not get_setting('scopefilter'):
            return False

        current_prefixes = list(get_setting('scopefilter.prefixes', []))
        current_asns = list(get_setting('scopefilter.asns', []))
        current_exclusions = {
            name
            for name, settings in get_setting('sources', {}).items()
            if settings.get('scopefilter_excluded')
        }

        if any([
            self.previous_scopefilter_prefixes != current_prefixes,
            self.previous_scopefilter_asns != current_asns,
            self.previous_scopefilter_excluded != current_exclusions,
        ]):
            self.previous_scopefilter_prefixes = current_prefixes
            self.previous_scopefilter_asns = current_asns
            self.previous_scopefilter_excluded = current_exclusions
            return True
        return False

    def run_if_relevant(self, source: str, runner_class, timer: int) -> bool:
        process_name = f'{runner_class.__name__}-{source}'

        current_time = time.time()
        has_expired = (self.last_started_time[process_name] + timer) < current_time
        if not has_expired or process_name in self.processes:
            return False

        logger.debug(f'Started new process {process_name} for mirror import/export for {source}')
        initiator = runner_class(source=source)
        process = ScheduledTaskProcess(runner=initiator, name=process_name)
        self.processes[process_name] = process
        process.start()
        self.last_started_time[process_name] = int(current_time)
        return True

    def terminate_children(self) -> None:  # pragma: no cover
        logger.info('MirrorScheduler terminating children')
        for process in self.processes.values():
            try:
                process.terminate()
                process.join()
            except Exception:
                pass

    def update_process_state(self):
        multiprocessing.active_children()  # to reap zombies
        for process_name, process in list(self.processes.items()):
            if process.is_alive():
                continue
            try:
                process.close()
            except Exception as e:  # pragma: no cover
                logging.error(f'Failed to close {process_name} (pid {process.pid}), '
                              f'possible resource leak: {e}')
            del self.processes[process_name]



def flatten_mock_calls(mock, flatten_objects=False):
    """
    Flatten the calls performed on a particular mock object,
    into a list of calls with arguments.

    If flatten_objects is set to True, objects of classes not in
    retained_classes are converted to strings.
    """
    result = []
    retained_classes = (int, list, tuple, set, bytes, bytearray)

    for call in mock.mock_calls:
        call = list(call)
        call_name = call[0]
        if '.' in str(call_name):
            call_name = str(call_name).split('.')[-1]
        if flatten_objects:
            args = tuple([str(a) if not isinstance(a, retained_classes) else a for a in call[1]])
        else:
            args = call[1]
        kwargs = call[2]
        result.append([call_name, args, kwargs])
    return result



import functools
import logging
from dataclasses import dataclass, field
from typing import Set, Tuple, List, Optional, TYPE_CHECKING

from ordered_set import OrderedSet
from passlib.hash import md5_crypt

from irrd.conf import get_setting
from irrd.rpsl.parser import RPSLObject
from irrd.rpsl.rpsl_objects import RPSLMntner, rpsl_object_from_text
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.queries import RPSLDatabaseQuery
from .parser_state import UpdateRequestType

if TYPE_CHECKING:  # pragma: no cover
    # http://mypy.readthedocs.io/en/latest/common_issues.html#import-cycles
    from .parser import ChangeRequest  # noqa: F401

logger = logging.getLogger(__name__)


@dataclass
class ValidatorResult:
    # OrderedSet has some obscure issues with mypy
    error_messages: Set[str] = field(default_factory=OrderedSet)  # type: ignore
    info_messages: Set[str] = field(default_factory=OrderedSet)  # type: ignore
    # mntners that may need to be notified
    mntners_notify: List[RPSLMntner] = field(default_factory=list)
    # whether the authentication succeeded due to use of an override password
    used_override: bool = field(default=False)

    def is_valid(self):
        return len(self.error_messages) == 0


class ReferenceValidator:
    """
    The ReferenceValidator validates references to other objects, given
    their expected object_class, source and PK.

    Sometimes updates are made to objects, referencing objects newly created
    in the same update message. To handle this, the validator can be preloaded
    with objects that should be considered valid.
    """
    def __init__(self, database_handler: DatabaseHandler) -> None:
        self.database_handler = database_handler
        self._cache: Set[Tuple[str, str, str]] = set()
        self._preloaded_new: Set[Tuple[str, str, str]] = set()
        self._preloaded_deleted: Set[Tuple[str, str, str]] = set()

    def preload(self, results: List['ChangeRequest']) -> None:
        """Preload an iterable of ChangeRequest objects to be considered valid, or to be considered deleted."""
        self._preloaded_new = set()
        self._preloaded_deleted = set()
        for request in results:
            assert request.rpsl_obj_new
            if request.request_type == UpdateRequestType.DELETE:
                self._preloaded_deleted.add((request.rpsl_obj_new.rpsl_object_class, request.rpsl_obj_new.pk(),
                                             request.rpsl_obj_new.source()))
            else:
                self._preloaded_new.add((request.rpsl_obj_new.rpsl_object_class, request.rpsl_obj_new.pk(),
                                         request.rpsl_obj_new.source()))

    def check_references_to_others(self, rpsl_obj: RPSLObject) -> ValidatorResult:
        """
        Check the validity of references of a particular object, i.e. whether
        all references to other objects actually exist in the database.
        """
        result = ValidatorResult()
        references = rpsl_obj.referred_strong_objects()
        source = rpsl_obj.source()

        for field_name, objects_referred, object_pks in references:
            for object_pk in object_pks:
                if not self._check_reference_to_others(objects_referred, object_pk, source):
                    if len(objects_referred) > 1:
                        objects_referred_str = 'one of ' + ', '.join(objects_referred)
                    else:
                        objects_referred_str = objects_referred[0]
                    result.error_messages.add(f'Object {object_pk} referenced in field {field_name} not found in '
                                              f'database {source} - must reference {objects_referred_str}.')
        return result

    def _check_reference_to_others(self, object_classes: List[str], object_pk: str, source: str) -> bool:
        """
        Check whether one reference to a particular object class/source/PK is valid,
        i.e. such an object exists in the database.

        Object classes is a list of classes to which the reference may point, e.g.
        person/role for admin-c, route for route-set members, or route/route6 for mp-members.
        """
        for object_class in object_classes:
            if (object_class, object_pk, source) in self._cache:
                return True
            if (object_class, object_pk, source) in self._preloaded_new:
                return True
            if (object_class, object_pk, source) in self._preloaded_deleted:
                return False

        query = RPSLDatabaseQuery().sources([source]).object_classes(object_classes).rpsl_pk(object_pk)
        results = list(self.database_handler.execute_query(query))
        for result in results:
            self._cache.add((result['object_class'], object_pk, source))
        if len(results):
            return True

        return False

    def check_references_from_others(self, rpsl_obj: RPSLObject) -> ValidatorResult:
        """
        Check for any references to this object in the DB.
        Used for validating deletions.

        Checks self._preloaded_deleted, because a reference from an object
        that is also about to be deleted, is acceptable.
        """
        result = ValidatorResult()
        if not rpsl_obj.references_strong_inbound():
            return result

        query = RPSLDatabaseQuery().sources([rpsl_obj.source()])
        query = query.lookup_attrs_in(rpsl_obj.references_strong_inbound(), [rpsl_obj.pk()])
        query_results = self.database_handler.execute_query(query)
        for query_result in query_results:
            reference_to_be_deleted = (query_result['object_class'], query_result['rpsl_pk'],
                                       query_result['source']) in self._preloaded_deleted
            if not reference_to_be_deleted:
                result.error_messages.add(f'Object {rpsl_obj.pk()} to be deleted, but still referenced '
                                          f'by {query_result["object_class"]} {query_result["rpsl_pk"]}')
        return result


class AuthValidator:
    """
    The AuthValidator validates authentication. It looks for relevant mntner
    objects and checks whether any of their auth methods pass.

    When adding a mntner in an update, a check for that mntner in the DB will
    fail, as it does not exist yet. To prevent this failure, call pre_approve()
    with a list of UpdateRequests.
    """
    passwords: List[str]
    overrides: List[str]
    keycert_obj_pk: Optional[str] = None

    def __init__(self, database_handler: DatabaseHandler, keycert_obj_pk=None) -> None:
        self.database_handler = database_handler
        self._mntner_db_cache: Set[RPSLMntner] = set()
        self._pre_approved: Set[str] = set()
        self.keycert_obj_pk = keycert_obj_pk

    def pre_approve(self, results: List['ChangeRequest']) -> None:
        """
        Pre-approve certain maintainers that are part of this batch of updates.
        This is required for creating new maintainers along with other objects.

        All new maintainer PKs are added to self._pre_approved. When they are
        encountered as mnt-by, the authentication is immediately approved,
        as a check in the database would fail.
        When the new mntner object's mnt-by is checked, there is an additional
        check to verify that it passes the newly submitted authentication.
        """
        self._pre_approved = set()
        for request in results:
            if request.is_valid() and request.request_type == UpdateRequestType.CREATE and isinstance(request.rpsl_obj_new, RPSLMntner):
                self._pre_approved.add(request.rpsl_obj_new.pk())

    def process_auth(self, rpsl_obj_new: RPSLObject, rpsl_obj_current: Optional[RPSLObject]) -> ValidatorResult:
        """
        Check whether authentication passes for all required objects.
        Returns a ValidatorResult object with error/info messages, and fills
        result.mntners_notify with the RPSLMntner objects that may have
        to be notified.

        If a valid override password is provided, changes are immediately approved.
        On the result object, used_override is set to True, but mntners_notify is
        not filled, as mntner resolving does not take place.
        """
        source = rpsl_obj_new.source()
        result = ValidatorResult()

        override_hash = get_setting('auth.override_password')
        if override_hash:
            for override in self.overrides:
                try:
                    if md5_crypt.verify(override, override_hash):
                        result.used_override = True
                        logger.debug('Found valid override password.')
                        return result
                    else:
                        logger.info('Found invalid override password, ignoring.')
                except ValueError as ve:
                    logger.error(f'Exception occurred while checking override password: {ve} (possible misconfigured hash?)')
        elif self.overrides:
            logger.info('Ignoring override password, auth.override_password not set.')

        mntners_new = rpsl_obj_new.parsed_data['mnt-by']
        logger.debug(f'Checking auth for new object {rpsl_obj_new}, mntners in new object: {mntners_new}')
        valid, mntner_objs_new = self._check_mntners(mntners_new, source)
        if not valid:
            self._generate_failure_message(result, mntners_new, rpsl_obj_new)

        if rpsl_obj_current:
            mntners_current = rpsl_obj_current.parsed_data['mnt-by']
            logger.debug(f'Checking auth for current object {rpsl_obj_current}, '
                         f'mntners in new object: {mntners_current}')
            valid, mntner_objs_current = self._check_mntners(mntners_current, source)
            if not valid:
                self._generate_failure_message(result, mntners_current, rpsl_obj_new)

            result.mntners_notify = mntner_objs_current
        else:
            result.mntners_notify = mntner_objs_new
            if get_setting('auth.authenticate_related_mntners'):
                mntners_related = self._find_related_mntners(rpsl_obj_new)
                if mntners_related:
                    related_object_class, related_pk, related_mntner_list = mntners_related
                    logger.debug(f'Checking auth for related object {related_object_class} / '
                                 f'{related_pk} with mntners {related_mntner_list}')
                    valid, mntner_objs_related = self._check_mntners(related_mntner_list, source)
                    if not valid:
                        self._generate_failure_message(result, related_mntner_list, rpsl_obj_new,
                                                       related_object_class, related_pk)
                        result.mntners_notify = mntner_objs_related

        if isinstance(rpsl_obj_new, RPSLMntner):
            if not rpsl_obj_current:
                result.error_messages.add('New mntner objects must be added by an administrator.')
                return result
            # Dummy auth values are only permitted in existing objects
            if rpsl_obj_new.has_dummy_auth_value():
                if len(self.passwords) == 1:
                    logger.debug(f'Object {rpsl_obj_new} submitted with dummy hash values and single password, '
                                 f'replacing all hashes with currently supplied password.')
                    rpsl_obj_new.force_single_new_password(self.passwords[0])
                    result.info_messages.add('As you submitted dummy hash values, all password hashes on this object '
                                             'were replaced with a new MD5-PW hash of the password you provided for '
                                             'authentication.')
                else:
                    result.error_messages.add('Object submitted with dummy hash values, but multiple or no passwords '
                                              'submitted. Either submit only full hashes, or a single password.')
            elif not rpsl_obj_new.verify_auth(self.passwords, self.keycert_obj_pk):
                result.error_messages.add('Authorisation failed for the auth methods on this mntner object.')

        return result

    def _check_mntners(self, mntner_pk_list: List[str], source: str) -> Tuple[bool, List[RPSLMntner]]:
        """
        Check whether authentication passes for a list of maintainers.

        Returns True if at least one of the mntners in mntner_list
        passes authentication, given self.passwords and
        self.keycert_obj_pk. Updates and checks self._mntner_db_cache
        to prevent double checking of maintainers.
        """
        mntner_pk_set = set(mntner_pk_list)
        mntner_objs: List[RPSLMntner] = [
            m for m in self._mntner_db_cache
            if m.pk() in mntner_pk_set and m.source() == source
        ]
        mntner_pks_to_resolve: Set[str] = mntner_pk_set - {m.pk() for m in mntner_objs}

        if mntner_pks_to_resolve:
            query = RPSLDatabaseQuery().sources([source])
            query = query.object_classes(['mntner']).rpsl_pks(mntner_pks_to_resolve)
            results = self.database_handler.execute_query(query)

            retrieved_mntner_objs: List[RPSLMntner] = [rpsl_object_from_text(r['object_text']) for r in results]   # type: ignore
            self._mntner_db_cache.update(retrieved_mntner_objs)
            mntner_objs += retrieved_mntner_objs

        for mntner_name in mntner_pk_list:
            if mntner_name in self._pre_approved:
                return True, mntner_objs

        for mntner_obj in mntner_objs:
            if mntner_obj.verify_auth(self.passwords, self.keycert_obj_pk):
                return True, mntner_objs

        return False, mntner_objs

    def _generate_failure_message(self, result: ValidatorResult, failed_mntner_list: List[str],
                                  rpsl_obj, related_object_class: Optional[str]=None,
                                  related_pk: Optional[str]=None) -> None:
        mntner_str = ', '.join(failed_mntner_list)
        msg = f'Authorisation for {rpsl_obj.rpsl_object_class} {rpsl_obj.pk()} failed: '
        msg += f'must by authenticated by one of: {mntner_str}'
        if related_object_class and related_pk:
            msg += f' - from parent {related_object_class} {related_pk}'
        result.error_messages.add(msg)

    @functools.lru_cache(maxsize=50)
    def _find_related_mntners(self, rpsl_obj_new: RPSLObject) -> Optional[Tuple[str, str, List[str]]]:
        """
        Find the maintainers of the related object to rpsl_obj_new, if any.
        This is used to authorise creating objects.

        Returns a tuple of:
        - object class of the related object
        - PK of the related object
        - List of maintainers for the related object (at least one must pass)
        Returns None of no related objects were found that should be authenticated.
        """
        if rpsl_obj_new.rpsl_object_class not in ['route', 'route6']:
            return None

        inetnum_class = {
            'route': 'inetnum',
            'route6': 'inet6num',
        }

        def init_query(rpsl_object_class: str) -> RPSLDatabaseQuery:
            query = RPSLDatabaseQuery().sources([rpsl_obj_new.source()])
            query = query.object_classes([rpsl_object_class])
            return query.first_only()

        object_class = inetnum_class[rpsl_obj_new.rpsl_object_class]
        query = init_query(object_class).ip_exact(rpsl_obj_new.prefix)
        inetnums = list(self.database_handler.execute_query(query))
        if not inetnums:
            query = init_query(object_class).ip_less_specific_one_level(rpsl_obj_new.prefix)
            inetnums = list(self.database_handler.execute_query(query))

        if inetnums:
            mntners = inetnums[0].get('parsed_data', {}).get('mnt-by', [])
            return inetnums[0]['object_class'], inetnums[0]['rpsl_pk'], mntners

        object_class = rpsl_obj_new.rpsl_object_class
        query = init_query(object_class).ip_less_specific_one_level(rpsl_obj_new.prefix)
        routes = list(self.database_handler.execute_query(query))
        if routes:
            mntners = routes[0].get('parsed_data', {}).get('mnt-by', [])
            return routes[0]['object_class'], routes[0]['rpsl_pk'], mntners

        return None



import os

import gzip
import logging
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile

from irrd.conf import get_setting
from irrd.rpki.status import RPKIStatus
from irrd.scopefilter.status import ScopeFilterStatus
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.queries import RPSLDatabaseQuery, DatabaseStatusQuery
from irrd.utils.text import remove_auth_hashes as remove_auth_hashes_func

EXPORT_PERMISSIONS = 0o644

logger = logging.getLogger(__name__)


class SourceExportRunner:
    """
    This SourceExportRunner is the entry point for the expect process
    for a single source.

    A gzipped file will be created in the export_destination directory
    with the contents of the source, along with a CURRENTSERIAL file.

    The contents of the source are first written to a temporary file, and
    then moved in place.
    """

    def __init__(self, source: str) -> None:
        self.source = source

    def run(self) -> None:
        self.database_handler = DatabaseHandler()
        try:
            export_destination = get_setting(f'sources.{self.source}.export_destination')
            if export_destination:
                logger.info(f'Starting a source export for {self.source} to {export_destination}')
                self._export(export_destination)

            export_destination_unfiltered = get_setting(f'sources.{self.source}.export_destination_unfiltered')
            if export_destination_unfiltered:
                logger.info(f'Starting an unfiltered source export for {self.source} '
                            f'to {export_destination_unfiltered}')
                self._export(export_destination_unfiltered, remove_auth_hashes=False)

            self.database_handler.commit()
        except Exception as exc:
            logger.error(f'An exception occurred while attempting to run an export '
                         f'for {self.source}: {exc}', exc_info=exc)
        finally:
            self.database_handler.close()

    def _export(self, export_destination, remove_auth_hashes=True):
        filename_export = Path(export_destination) / f'{self.source.lower()}.db.gz'
        export_tmpfile = NamedTemporaryFile(delete=False)
        filename_serial = Path(export_destination) / f'{self.source.upper()}.CURRENTSERIAL'

        query = DatabaseStatusQuery().source(self.source)

        try:
            serial = next(self.database_handler.execute_query(query))['serial_newest_seen']
        except StopIteration:
            serial = None

        with gzip.open(export_tmpfile.name, 'wb') as fh:
            query = RPSLDatabaseQuery().sources([self.source])
            query = query.rpki_status([RPKIStatus.not_found, RPKIStatus.valid])
            query = query.scopefilter_status([ScopeFilterStatus.in_scope])
            for obj in self.database_handler.execute_query(query):
                object_text = obj['object_text']
                if remove_auth_hashes:
                    object_text = remove_auth_hashes_func(object_text)
                object_bytes = object_text.encode('utf-8')
                fh.write(object_bytes + b'\n')
            fh.write(b'# EOF\n')

        os.chmod(export_tmpfile.name, EXPORT_PERMISSIONS)
        if filename_export.exists():
            os.unlink(filename_export)
        if filename_serial.exists():
            os.unlink(filename_serial)
        shutil.move(export_tmpfile.name, filename_export)

        if serial is not None:
            with open(filename_serial, 'w') as fh:
                fh.write(str(serial))
            os.chmod(filename_serial, EXPORT_PERMISSIONS)

        self.database_handler.record_serial_exported(self.source, serial)
        logger.info(f'Export for {self.source} complete at serial {serial}, stored in {filename_export} / {filename_serial}')



#!/usr/bin/env python
# flake8: noqa: E402

"""
A simple load tester for IRRd.
Sends random !g queries.
"""
import time

import argparse
import random
import socket


def main(host, port, count):
    queries = [b'!!\n']
    for i in range(count):
        asn = random.randrange(1, 50000)
        query = f'!gAS{asn}\n'.encode('ascii')
        queries.append(query)
    queries.append(b'!q\n')

    s = socket.socket()
    s.settimeout(600)
    s.connect((host, port))

    queries_str = b''.join(queries)
    s.sendall(queries_str)

    start_time = time.perf_counter()
    while 1:
        data = s.recv(1024*1024)
        if not data:
            break

    elapsed = time.perf_counter() - start_time
    time_per_query = elapsed / count * 1000
    qps = int(count / elapsed)
    print(f'Ran {count} queries in {elapsed}s, time per query {time_per_query} ms, {qps} qps')


if __name__ == '__main__':  # pragma: no cover
    description = """A simple load tester for IRRd. Sends random !g queries."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--count', dest='count', type=int, default=5000,
                        help=f'number of queries to run (default: 5000)')
    parser.add_argument('host', type=str,
                        help='hostname of instance')
    parser.add_argument('port', type=int,
                        help='port of instance')
    args = parser.parse_args()

    main(args.host, args.port, args.count)



import ctypes
import logging
import os
import signal
from multiprocessing import Process

logger = logging.getLogger(__name__)


# Covered by integration tests


class ExceptionLoggingProcess(Process):  # pragma: no cover
    def run(self) -> None:
        try:
            super().run()
        except Exception as e:
            logger.critical(f'Essential IRRd subprocess encountered a fatal error, '
                            f'traceback follows, shutting down: {e}', exc_info=e)
            os.kill(os.getppid(), signal.SIGTERM)


def memory_trim():  # pragma: no cover
    # See https://github.com/irrdnet/irrd/issues/571
    try:
        ctypes.CDLL(None).malloc_trim(0)
    except Exception:
        pass



import importlib.util
import sys
import time

import logging.config
import os
import re
import signal
import yaml
from IPy import IP
from dotted.collection import DottedDict
from pathlib import Path
from typing import Any, List, Optional

CONFIG_PATH_DEFAULT = '/etc/irrd.yaml'

logger = logging.getLogger(__name__)
PASSWORD_HASH_DUMMY_VALUE = 'DummyValue'
SOURCE_NAME_RE = re.compile('^[A-Z][A-Z0-9-]*[A-Z0-9]$')
RPKI_IRR_PSEUDO_SOURCE = 'RPKI'


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s irrd[%(process)d]: [%(name)s#%(levelname)s] %(message)s'
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        },
    },
    'loggers': {
        # Tune down some very loud and not very useful loggers from libraries.
        'passlib.registry': {
            'level': 'INFO',
        },
        'gnupg': {
            'level': 'INFO',
        },
        'sqlalchemy': {
            'level': 'WARNING',
        },
        '': {
            'handlers': ['console'],
            'level': 'INFO',
        },
    }
}


logging.config.dictConfig(LOGGING)
logging.Formatter.converter = time.gmtime


class ConfigurationError(ValueError):
    pass


# testing_overrides can be set to a DottedDict, and is used
# to override settings while testing, using the config_override
# fixture.
testing_overrides = None


class Configuration:
    """
    The Configuration class stores the current IRRD configuration,
    checks the validity of the settings, and offers graceful reloads.
    """
    user_config_staging: DottedDict
    user_config_live: DottedDict

    def __init__(self, user_config_path: Optional[str]=None, commit=True):
        """
        Load the default config and load and check the user provided config.
        If a logfile was specified, direct logs there.
        """
        self.user_config_path = user_config_path if user_config_path else CONFIG_PATH_DEFAULT
        default_config_path = str(Path(__file__).resolve().parents[0] / 'default_config.yaml')
        default_config_yaml = yaml.safe_load(open(default_config_path))
        self.default_config = DottedDict(default_config_yaml['irrd'])

        errors = self._staging_reload_check(log_success=False)
        if errors:
            raise ConfigurationError(f'Errors found in configuration, unable to start: {errors}')

        if commit:
            self._commit_staging()

            logging_config_path = self.get_setting_live('log.logging_config_path')
            logfile_path = self.get_setting_live('log.logfile_path')
            if logging_config_path:
                spec = importlib.util.spec_from_file_location("logging_config", logging_config_path)
                config_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(config_module)  # type: ignore
                self.logging_config = config_module.LOGGING  # type: ignore
                logging.config.dictConfig(self.logging_config)
            elif logfile_path:
                LOGGING['handlers']['file'] = {   # type:ignore
                    'class': 'logging.handlers.WatchedFileHandler',
                    'filename': logfile_path,
                    'formatter': 'verbose',
                }
                # noinspection PyTypeChecker
                LOGGING['loggers']['']['handlers'] = ['file']   # type:ignore
                logging.config.dictConfig(LOGGING)

            # Re-commit to apply loglevel
            self._commit_staging()

    def get_setting_live(self, setting_name: str, default: Any=None) -> Any:
        """
        Get a setting from the live config.
        In order, this will look in:
        - A env variable, uppercase and dots replaced by underscores, e.g.
          IRRD_SERVER_WHOIS_INTERFACE
        - The testing_overrides DottedDict
        - The live user config.
        - The default config.

        If it is not found in any, the value of the default paramater
        is returned, which is None by default.
        """
        env_key = 'IRRD_' + setting_name.upper().replace('.', '_')
        if env_key in os.environ:
            return os.environ[env_key]
        if testing_overrides:
            try:
                return testing_overrides[setting_name]
            except KeyError:
                pass
        try:
            return self.user_config_live[setting_name]
        except KeyError:
            return self.default_config.get(setting_name, default)

    def reload(self) -> bool:
        """
        Reload the configuration, if it passes the checks.
        """
        errors = self._staging_reload_check()
        if errors:
            logger.error(f'Errors found in configuration, continuing with current settings: {errors}')
            return False

        self._commit_staging()
        return True

    def _commit_staging(self) -> None:
        """
        Activate the current staging config as the live config.
        """
        self.user_config_live = self.user_config_staging
        logging.getLogger('').setLevel(self.get_setting_live('log.level', default='INFO'))
        if hasattr(sys, '_called_from_test'):
            logging.getLogger('').setLevel('DEBUG')

    def _staging_reload_check(self, log_success=True) -> List[str]:
        """
        Reload the staging configuration, and run the config checks on it.
        Returns a list of errors if any were found, or an empty list of the
        staging config is valid.
        """
        # While in testing, Configuration does not demand a valid config file
        # This simplifies test setup, as most tests do not need it.
        # If a non-default path is set during testing, it is still checked.
        if hasattr(sys, '_called_from_test') and self.user_config_path == CONFIG_PATH_DEFAULT:
            self.user_config_staging = DottedDict({})
            return []

        try:
            with open(self.user_config_path) as fh:
                user_config_yaml = yaml.safe_load(fh)
        except OSError as oe:
            return [f'Error opening config file {self.user_config_path}: {oe}']
        except yaml.YAMLError as ye:
            return [f'Error parsing YAML file: {ye}']

        if not isinstance(user_config_yaml, dict) or 'irrd' not in user_config_yaml:
            return [f'Could not find root item "irrd" in config file {self.user_config_path}']
        self.user_config_staging = DottedDict(user_config_yaml['irrd'])

        errors = self._check_staging_config()
        if not errors and log_success:
            logger.info(f'Configuration successfully (re)loaded from {self.user_config_path} in PID {os.getpid()}')
        return errors

    def _check_staging_config(self) -> List[str]:
        """
        Validate the current staging configuration.
        Returns a list of any errors, or an empty list for a valid config.
        """
        errors = []
        config = self.user_config_staging

        if not self._check_is_str(config, 'database_url'):
            errors.append('Setting database_url is required.')

        if not self._check_is_str(config, 'redis_url'):
            errors.append('Setting redis_url is required.')

        if not self._check_is_str(config, 'piddir') or not os.path.isdir(config['piddir']):
            errors.append('Setting piddir is required and must point to an existing directory.')

        expected_access_lists = {
            config.get('server.whois.access_list'),
            config.get('server.http.access_list'),
        }

        if not self._check_is_str(config, 'email.from') or '@' not in config.get('email.from'):
            errors.append('Setting email.from is required and must be an email address.')
        if not self._check_is_str(config, 'email.smtp'):
            errors.append('Setting email.smtp is required.')
        if not self._check_is_str(config, 'email.recipient_override', required=False) \
                or '@' not in config.get('email.recipient_override', '@'):
            errors.append('Setting email.recipient_override must be an email address if set.')

        string_not_required = ['email.footer', 'server.whois.access_list', 'server.http.access_list',
                               'rpki.notify_invalid_subject', 'rpki.notify_invalid_header',
                               'rpki.slurm_source']
        for setting in string_not_required:
            if not self._check_is_str(config, setting, required=False):
                errors.append(f'Setting {setting} must be a string, if defined.')

        if not self._check_is_str(config, 'auth.gnupg_keyring'):
            errors.append('Setting auth.gnupg_keyring is required.')

        for name, access_list in config.get('access_lists', {}).items():
            for item in access_list:
                try:
                    IP(item)
                except ValueError as ve:
                    errors.append(f'Invalid item in access list {name}: {ve}.')

        for prefix in config.get('scopefilter.prefixes', []):
            try:
                IP(prefix)
            except ValueError as ve:
                errors.append(f'Invalid item in prefix scopefilter: {prefix}: {ve}.')

        for asn in config.get('scopefilter.asns', []):
            try:
                if '-' in str(asn):
                    first, last = asn.split('-')
                    int(first)
                    int(last)
                else:
                    int(asn)
            except ValueError:
                errors.append(f'Invalid item in asn scopefilter: {asn}.')

        known_sources = set(config.get('sources', {}).keys())

        has_authoritative_sources = False
        for name, details in config.get('sources', {}).items():
            if details.get('authoritative'):
                has_authoritative_sources = True
            if config.get('rpki.roa_source') and name == RPKI_IRR_PSEUDO_SOURCE:
                errors.append(f'Setting sources contains reserved source name: {RPKI_IRR_PSEUDO_SOURCE}')
            if not SOURCE_NAME_RE.match(name):
                errors.append(f'Invalid source name: {name}')

            nrtm_mirror = details.get('nrtm_host') and details.get('import_serial_source')
            if details.get('keep_journal') and not (nrtm_mirror or details.get('authoritative')):
                errors.append(f'Setting keep_journal for source {name} can not be enabled unless either authoritative '
                              f'is enabled, or all three of nrtm_host, nrtm_port and import_serial_source.')
            if details.get('nrtm_host') and not details.get('import_serial_source'):
                errors.append(f'Setting nrtm_host for source {name} can not be enabled without setting '
                              f'import_serial_source.')

            if details.get('authoritative') and (details.get('nrtm_host') or details.get('import_source')):
                errors.append(f'Setting authoritative for source {name} can not be enabled when either '
                              f'nrtm_host or import_source are set.')

            if config.get('database_readonly') and (details.get('authoritative') or details.get('nrtm_host') or details.get('import_source')):
                errors.append(f'Source {name} can not have authoritative, import_source or nrtm_host set '
                              f'when database_readonly is enabled.')

            if not str(details.get('nrtm_port', '43')).isnumeric():
                errors.append(f'Setting nrtm_port for source {name} must be a number.')
            if not str(details.get('import_timer', '0')).isnumeric():
                errors.append(f'Setting import_timer for source {name} must be a number.')
            if not str(details.get('export_timer', '0')).isnumeric():
                errors.append(f'Setting export_timer for source {name} must be a number.')

            if details.get('nrtm_access_list'):
                expected_access_lists.add(details.get('nrtm_access_list'))
            if details.get('nrtm_access_list_unfiltered'):
                expected_access_lists.add(details.get('nrtm_access_list_unfiltered'))

        if config.get('rpki.roa_source', 'https://rpki.gin.ntt.net/api/export.json'):
            known_sources.add(RPKI_IRR_PSEUDO_SOURCE)
            if has_authoritative_sources and config.get('rpki.notify_invalid_enabled') is None:
                errors.append('RPKI-aware mode is enabled and authoritative sources are configured, '
                              'but rpki.notify_invalid_enabled is not set. Set to true or false.'
                              'DANGER: care is required with this setting in testing setups with '
                              'live data, as it may send bulk emails to real resource contacts '
                              'unless email.recipient_override is also set. '
                              'Read documentation carefully.')

        unknown_default_sources = set(config.get('sources_default', [])).difference(known_sources)
        if unknown_default_sources:
            errors.append(f'Setting sources_default contains unknown sources: {", ".join(unknown_default_sources)}')

        if not str(config.get('rpki.roa_import_timer', '0')).isnumeric():
            errors.append('Setting rpki.roa_import_timer must be set to a number.')

        if config.get('log.level') and not config.get('log.level') in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
            errors.append(f'Invalid log.level: {config.get("log.level")}. '
                          f'Valid settings for log.level are `DEBUG`, `INFO`, `WARNING`, `ERROR`, `CRITICAL`.')
        if config.get('log.logging_config_path') and (config.get('log.logfile_path') or config.get('log.level')):
            errors.append('Setting log.logging_config_path can not be combined with'
                          'log.logfile_path or log.level')

        access_lists = set(config.get('access_lists', {}).keys())
        unresolved_access_lists = [x for x in expected_access_lists.difference(access_lists) if x and isinstance(x, str)]
        unresolved_access_lists.sort()
        if unresolved_access_lists:
            errors.append(f'Access lists {", ".join(unresolved_access_lists)} referenced in settings, but not defined.')

        return errors

    def _check_is_str(self, config, key, required=True):
        if required:
            return config.get(key) and isinstance(config.get(key), str)
        return config.get(key) is None or isinstance(config.get(key), str)


configuration = None


def get_configuration() -> Optional[Configuration]:
    """
    Get the Configuration object, if initialised.
    """
    global configuration
    return configuration


def config_init(config_path, commit=True) -> None:
    """
    Initialise the configuration from a configuration path.
    If commit is False, only validates the configuration.
    """
    global configuration
    configuration = Configuration(config_path, commit)


def is_config_initialised() -> bool:
    """
    Returns whether the configuration is initialised,
    i.e. whether get_setting() can be called.
    """
    configuration = get_configuration()
    return configuration is not None


def get_setting(setting_name: str, default: Any=None) -> Any:
    """
    Convenience wrapper to get the value of a setting.
    """
    configuration = get_configuration()
    if not configuration:  # pragma: no cover
        raise Exception('get_setting() called before configuration was initialised')
    return configuration.get_setting_live(setting_name, default)


def sighup_handler(signum, frame) -> None:
    """
    Reload the settings when a SIGHUP is received.
    Note that not all processes re-read their settings on every run,
    so not all settings can be changed while running.
    """
    configuration = get_configuration()
    if configuration:
        configuration.reload()


signal.signal(signal.SIGHUP, sighup_handler)



#!/usr/bin/env python
# flake8: noqa: E402
import argparse
import logging
import sys

from pathlib import Path

"""
Load an RPSL file into the database.
"""

logger = logging.getLogger(__name__)
sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.conf import config_init, CONFIG_PATH_DEFAULT, get_setting

from irrd.storage.database_handler import DatabaseHandler

def set_force_reload(source) -> None:
    dh = DatabaseHandler()
    dh.set_force_reload(source)
    dh.commit()
    dh.close()


def main():  # pragma: no cover
    description = """Force a full reload for a mirror."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    parser.add_argument('source', type=str,
                        help='the name of the source to reload')
    args = parser.parse_args()

    config_init(args.config_file_path)
    if get_setting('database_readonly'):
        print('Unable to run, because database_readonly is set')
        sys.exit(-1)

    set_force_reload(args.source)


if __name__ == '__main__':  # pragma: no cover
    main()



from typing import TypeVar, List

from IPy import IP

RPSLParserMessagesType = TypeVar('RPSLParserMessagesType', bound='RPSLParserMessages')


class RPSLParserMessages:
    levels = ['INFO', 'ERROR']

    def __init__(self) -> None:
        self._messages: List[tuple] = []

    def __str__(self) -> str:
        messages_str = [f'{msg[0]}: {msg[1]}' for msg in self._messages]
        return '\n'.join(messages_str)

    def messages(self) -> List[str]:
        return [msg[1] for msg in self._messages]

    def infos(self) -> List[str]:
        return [msg[1] for msg in self._messages if msg[0] == 'INFO']

    def errors(self) -> List[str]:
        return [msg[1] for msg in self._messages if msg[0] == 'ERROR']

    def info(self, msg: str) -> None:
        self._message('INFO', msg)

    def error(self, msg: str) -> None:
        self._message('ERROR', msg)

    def merge_messages(self, other_messages: RPSLParserMessagesType) -> None:
        self._messages += other_messages._messages

    def _message(self, level: str, message: str) -> None:
        self._messages.append((level, message))


class RPSLFieldParseResult:
    def __init__(self, value: str, values_list: List[str]=None, ip_first: IP=None, ip_last: IP=None,
                 prefix: IP=None, prefix_length: int=None, asn_first: int=None, asn_last: int=None) -> None:
        self.value = value
        self.values_list = values_list
        self.ip_first = ip_first
        self.ip_last = ip_last
        self.prefix = prefix
        self.prefix_length = prefix_length
        self.asn_first = asn_first
        self.asn_last = asn_last



"""Add bogon_status

Revision ID: 39e4f15ed80c
Revises: 1743f98a456d
Create Date: 2020-04-22 14:43:57.985437

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine.reflection import Inspector


# revision identifiers, used by Alembic.

revision = '39e4f15ed80c'
down_revision = '1743f98a456d'
branch_labels = None
depends_on = None


# This is a very strange migration.
# The bogon_status column was added at some point, but was never used.
# Some people's deployments have this migration already deployed, so
# this migration can't be removed. For anyone upgrading from earlier,
# we don't want to re-add it to then remove it in the next migration,
# because it takes a lot of time.
#
# Therefore, this migration does nothing to upgrade, and on downgrade
# removes the column, but only if it exists.
# It assumes the existence of the bogonstatus enum matches that of
# the bogon_status column.

def upgrade():
    pass


def downgrade():
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)
    columns = inspector.get_columns('rpsl_objects')

    if 'bogon_status' in columns:
        op.drop_index(op.f('ix_rpsl_objects_bogon_status'), table_name='rpsl_objects')
        op.drop_column('rpsl_objects', 'bogon_status')
        bogon_status = sa.Enum('unknown', 'not_bogon', 'bogon_as', 'bogon_prefix', name='bogonstatus')
        bogon_status.drop(op.get_bind())



import datetime
import re
from typing import List, Type, Optional

from IPy import IP

from .config import PASSWORD_HASHERS
from .parser_state import RPSLParserMessages, RPSLFieldParseResult
from irrd.utils.validators import parse_as_number, ValidationError

# The IPv4/IPv6 regexes are for initial screening - not full validators
re_ipv4_prefix = re.compile(r"^\d+\.\d+\.\d+\.\d+/\d+$")
re_ipv6_prefix = re.compile(r"^[A-F\d:]+/\d+$", re.IGNORECASE)

# This regex is not designed to catch every possible invalid variation,
# but rather meant to protect against unintentional mistakes.
#                         # Validate local-part           @ domain         | or IPv4 address        | or IPv6
re_email = re.compile(r"^[A-Z0-9$!#%&\"*+\/=?^_`{|}~\\.-]+@(([A-Z0-9\\.-]+)|(\[\d+\.\d+\.\d+\.\d+\])|(\[[A-f\d:]+\]))$", re.IGNORECASE)

re_range_operator = re.compile(r"^((\d{1,3}-\d{1,3})|(-)|(\+)|(\d{1,3}))$")
re_pgpkey = re.compile(r"^PGPKEY-[A-F0-9]{8}$")
re_dnsname = re.compile(r"^(([A-Z0-9]|[A-Z0-9][A-Z0-9\-]*[A-Z0-9])\.)*([A-Z0-9]|[A-Z0-9][A-Z0-9\-]*[A-Z0-9])$", re.IGNORECASE)
re_generic_name = re.compile(r"^[A-Z][A-Z0-9_-]*[A-Z0-9]$", re.IGNORECASE)
reserved_words = ["ANY", "AS-ANY", "RS_ANY", "PEERAS", "AND", "OR", "NOT", "ATOMIC", "FROM", "TO", "AT", "ACTION",
                  "ACCEPT", "ANNOUNCE", "EXCEPT", "REFINE", "NETWORKS", "INTO", "INBOUND", "OUTBOUND"]
reserved_prefixes = ["AS-", "RS-", "RTRS-", "FLTR-", "PRNG-"]

# Turn "IP('193.0.1.1/21') has invalid prefix length (21)" into "invalid prefix length (21)"
re_clean_ip_error = re.compile(r"IP\('[A-F0-9:./]+'\) has ", re.IGNORECASE)

"""
Fields for RPSL data.

Note that these objects are instantiated once per attribute during RPSL object
class loading. Therefore, the instances will be shared between different RPSL
objects. In other words, two role object's nic-hdl fields will use the same
instance of RPSLGenericNameField. Never store object-specific state in self
in any field instances.
"""


class RPSLTextField:
    """
    Base field class for simple RPSL text fields. All other fields should inherit from this class.

    Note that parse() can expect data to be stripped from whitespaces and comments, and multiple
    lines be joined by commas. It should add any info, error or warning messages to the passed
    messages object, and return a parsed version of the value. If it is not possible to extract
    a value, e.g. due to a validation error, it should return None.

    The keep_case property affects the generation of RPSLObject.parsed_data during parsing.
    If keep_case is False, all data is converted to upper case. The original object text is not
    modified for this. As parsed_data is indexed in the databases, this is important for searches,
    to match e.g. 'mntner: FOO' to 'mnt-by: foo' - as these values are equivalent.
    """
    keep_case = True

    def __init__(self, optional: bool=False, multiple: bool=False, primary_key: bool=False, lookup_key: bool=False) -> None:
        self.optional = optional
        self.multiple = multiple
        self.primary_key = primary_key
        self.lookup_key = lookup_key

    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        return RPSLFieldParseResult(value)


class RPSLFieldListMixin:
    """
    Mixin to allow fields to support list values, like 'AS1, AS2, AS3'.

    For example, if you have an RPSLASNumberField that validates a single
    AS number, you can create RPSLASNumbersField that allows a list of
    AS numbers by creating:

        class RPSLASNumbersField(RPSLFieldListMixin, RPSLASNumberField):
            pass
    """
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        parse_results = []
        for single_value in value.split(','):
            single_value = single_value.strip()
            if single_value:
                parse_result = super().parse(single_value, messages, strict_validation)  # type: ignore
                parse_results.append(parse_result)
        if not all(parse_results):
            return None
        values = [result.value for result in parse_results]
        return RPSLFieldParseResult(','.join(values), values_list=values)


class RPSLIPv4PrefixField(RPSLTextField):
    """Field for a single IPv4 prefix."""
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if not re_ipv4_prefix.match(value):
            messages.error(f'Invalid address prefix: {value}')
            return None

        try:
            ip = IP(value, ipversion=4)
        except ValueError as ve:
            clean_error = clean_ip_value_error(ve)
            messages.error(f'Invalid address prefix: {value}: {clean_error}')
            return None

        parsed_ip_str = str(ip)
        if ip.prefixlen() == 32:
            parsed_ip_str += '/32'
        if parsed_ip_str != value:
            messages.info(f'Address prefix {value} was reformatted as {parsed_ip_str}')
        return RPSLFieldParseResult(parsed_ip_str, ip_first=ip.net(), ip_last=ip.broadcast(),
                                    prefix=ip, prefix_length=ip.prefixlen())


class RPSLIPv4PrefixesField(RPSLFieldListMixin, RPSLIPv4PrefixField):
    """Field for a comma-separated list of IPv4 prefixes."""
    pass


class RPSLIPv6PrefixField(RPSLTextField):
    """Field for a single IPv6 prefix."""
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if not re_ipv6_prefix.match(value):
            messages.error(f'Invalid address prefix: {value}')
            return None

        try:
            ip = IP(value, ipversion=6)
        except ValueError as ve:
            clean_error = clean_ip_value_error(ve)
            messages.error(f'Invalid address prefix: {value}: {clean_error}')
            return None

        parsed_ip_str = str(ip)
        if ip.prefixlen() == 128:
            parsed_ip_str += '/128'
        if parsed_ip_str != value:
            messages.info(f'Address prefix {value} was reformatted as {parsed_ip_str}')
        return RPSLFieldParseResult(parsed_ip_str, ip_first=ip.net(), ip_last=ip.broadcast(),
                                    prefix=ip, prefix_length=ip.prefixlen())


class RPSLIPv6PrefixesField(RPSLFieldListMixin, RPSLIPv6PrefixField):
    """Field for a comma-separated list of IPv6 prefixes."""
    pass


class RPSLIPv4AddressRangeField(RPSLTextField):
    """
    Field for a range of IPv4 addresses, as used in inetnum keys.

    Note that a single IP address is also valid, and that the range does
    not have to align to bitwise boundaries of prefixes.
    """
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        value = value.replace(',', '')  # #311, process multiline PK correctly
        if '-' in value:
            ip1_input, ip2_input = value.split('-', 1)
        else:
            ip1_input = ip2_input = value

        try:
            ip1 = IP(ip1_input)
            ip2 = IP(ip2_input)
        except ValueError as ve:
            clean_error = clean_ip_value_error(ve)
            messages.error(f'Invalid address range: {value}: {clean_error}')
            return None

        if not ip1.version() == ip2.version() == 4:
            messages.error(f'Invalid address range: {value}: IP version mismatch')
            return None
        if ip1.int() > ip2.int():
            messages.error(f'Invalid address range: {value}: first IP is higher than second IP')
            return None

        if '-' in value:
            parsed_value = f'{ip1} - {ip2}'
        else:
            parsed_value = str(ip1)
        if parsed_value != value:
            messages.info(f'Address range {value} was reformatted as {parsed_value}')
        return RPSLFieldParseResult(parsed_value, ip_first=ip1, ip_last=ip2)


class RPSLRouteSetMemberField(RPSLTextField):
    """
    Field for the members of a route-set. These can be:
        - A valid name for another route set.
        - A valid IPv4 or IPv6 (depending on ip_version) prefix
        - A valid prefix followed by:
          - ^-
          - ^+
          - ^[integer]
          - ^[integer]-[integer]
    """
    keep_case = True

    def __init__(self, ip_version: Optional[int], *args, **kwargs) -> None:
        if ip_version and ip_version not in [4, 6]:
            raise ValueError(f'Invalid IP version: {ip_version}')
        self.ip_version = ip_version
        super().__init__(*args, **kwargs)

    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if '^' in value:
            address, range_operator = value.split('^', maxsplit=1)
            if not range_operator:
                messages.error(f'Missing range operator in value: {value}')
                return None
        else:
            address = value
            range_operator = ''

        parse_set_result_messages = RPSLParserMessages()
        parse_set_result = parse_set_name(['RS-', 'AS-'], address, parse_set_result_messages, strict_validation)
        if parse_set_result and not parse_set_result_messages.errors():
            result_value = parse_set_result.value
            if range_operator:
                result_value += '^' + range_operator
            if result_value != value:
                messages.info(f'Route set member {value} was reformatted as {result_value}')
            return RPSLFieldParseResult(value=result_value)
        try:
            parsed_str, parsed_int = parse_as_number(address)
            result_value = parsed_str
            if range_operator:
                result_value += '^' + range_operator
            if result_value != value:
                messages.info(f'Route set member {value} was reformatted as {result_value}')
            return RPSLFieldParseResult(value=result_value)
        except ValidationError:
            pass

        try:
            ip_version = self.ip_version if self.ip_version else 0
            ip = IP(address, ipversion=ip_version)
        except ValueError as ve:
            clean_error = clean_ip_value_error(ve)
            messages.error(f'Value is neither a valid set name nor a valid prefix: {address}: {clean_error}')
            return None

        if range_operator and not re_range_operator.match(range_operator):
            messages.error(f'Invalid range operator {range_operator} in value: {value}')
            return None

        parsed_ip_str = str(ip)
        if ip.version() == 4 and ip.prefixlen() == 32:
            parsed_ip_str += '/32'
        if ip.version() == 6 and ip.prefixlen() == 128:
            parsed_ip_str += '/128'
        if range_operator:
            parsed_ip_str += '^' + range_operator

        if parsed_ip_str != value:
            messages.info(f'Route set member {value} was reformatted as {parsed_ip_str}')
        return RPSLFieldParseResult(parsed_ip_str)


class RPSLRouteSetMembersField(RPSLFieldListMixin, RPSLRouteSetMemberField):
    pass


class RPSLASNumberField(RPSLTextField):
    """Field for a single AS number (in ASxxxx syntax)."""
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        try:
            parsed_str, parsed_int = parse_as_number(value)
        except ValidationError as ve:
            messages.error(str(ve))
            return None
        if parsed_str and parsed_str.upper() != value.upper():
            messages.info(f'AS number {value} was reformatted as {parsed_str}')
        return RPSLFieldParseResult(parsed_str, asn_first=parsed_int, asn_last=parsed_int)


class RPSLASBlockField(RPSLTextField):
    """Field for a block of AS numbers, e.g. AS1 - AS5."""
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if '-' not in value:
            messages.error(f'Invalid AS range: {value}: does not contain a hyphen')
            return None

        as1_raw, as2_raw = map(str.strip, value.split('-', 1))

        try:
            as1_str, as1_int = parse_as_number(as1_raw)
            as2_str, as2_int = parse_as_number(as2_raw)
        except ValidationError as ve:
            messages.error(str(ve))
            return None

        if as1_int > as2_int:  # type: ignore
            messages.error(f'Invalid AS range: {value}: first AS is higher then second AS')
            return None

        parsed_value = f'{as1_str} - {as2_str}'
        if parsed_value != value:
            messages.info(f'AS range {value} was reformatted as {parsed_value}')
        return RPSLFieldParseResult(parsed_value, asn_first=as1_int, asn_last=as2_int)


class RPSLSetNameField(RPSLTextField):
    """
    Field for set names, i.e. names of objects like route-set, prefix-set, etc.

    The actual set name must start with a designated prefix (which is otherwise not
    permitted for RPSL names).
    Set names can consist of multiple components, e.g. AS65537:RS-FOO. Each
    component must be a valid set name or valid AS number, and one component
    must be a valid set name for this specific set, i.e. start with the given prefix.

    The prefix provided is the expected prefix of the set name, e.g. 'RS' for
    a route-set, or 'AS' for an as-set.R
    """
    keep_case = False

    def __init__(self, prefix: str, *args, **kwargs) -> None:
        self.prefix = prefix + '-'
        super().__init__(*args, **kwargs)

    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        return parse_set_name([self.prefix], value, messages, strict_validation)


class RPSLEmailField(RPSLTextField):
    """Field for an e-mail address. Only performs basic validation."""
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if not re_email.match(value):
            messages.error(f'Invalid e-mail address: {value}')
            return None
        return RPSLFieldParseResult(value)


class RPSLChangedField(RPSLTextField):
    """Field for an changed line. Only performs basic validation for email."""
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        date: Optional[str]
        try:
            email, date = value.split(' ')
        except ValueError:
            email = value
            date = None

        if not re_email.match(email):
            messages.error(f'Invalid e-mail address: {email}')
            return None
        if date:
            try:
                datetime.datetime.strptime(date, '%Y%m%d')
            except ValueError as ve:
                messages.error(f'Invalid changed date: {date}: {ve}')
                return None
        return RPSLFieldParseResult(value)


class RPSLDNSNameField(RPSLTextField):
    """Field for a DNS name, as used in e.g. inet-rtr names."""
    keep_case = False

    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if not re_dnsname.match(value):
            messages.error(f'Invalid DNS name: {value}')
            return None
        return RPSLFieldParseResult(value)


class RPSLGenericNameField(RPSLTextField):
    """
    Field for a generic name.

    Generic names are names of objects that do not have a more strict definition,
    such as the nic-hdl of a person.
    Optionally, a list of allowed reserved prefixes can be provided.
    For example, this is used for the as-name attribute of an aut-num, as they
    are allowed to start with 'AS'.

    If non_strict_allow_any is set, the parser will allow any value if strict_validation
    is disabled. This is needed on nic-hdl for legacy reasons -
    see https://github.com/irrdnet/irrd/issues/60
    """
    keep_case = False

    def __init__(self, allowed_prefixes: List[str]=None, non_strict_allow_any=False, *args, **kwargs) -> None:
        self.non_strict_allow_any = non_strict_allow_any
        if allowed_prefixes:
            self.allowed_prefixes = [prefix.upper() + '-' for prefix in allowed_prefixes]
        else:
            self.allowed_prefixes = []
        super().__init__(*args, **kwargs)

    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if not strict_validation and self.non_strict_allow_any:
            return RPSLFieldParseResult(value)

        upper_value = value.upper()
        if strict_validation:
            if upper_value in reserved_words:
                messages.error(f'Invalid name: {value}: this is a reserved word')
                return None

            for prefix in reserved_prefixes:
                if upper_value.startswith(prefix) and prefix not in self.allowed_prefixes:
                    messages.error(f'Invalid name: {value}: {prefix} is a reserved prefix')
                    return None

        if not re_generic_name.match(upper_value):
            messages.error(f'Invalid name: {value}: contains invalid characters, does not start with a letter, '
                           f'or does not end in a letter/digit')
            return None
        return RPSLFieldParseResult(value)


class RPSLReferenceField(RPSLTextField):
    """
    Field for a reference to another field.

    Example: the mntner field in a person object, refers to the primary key
    field of the mntner object. Upon validating a person:mntner value, this
    field will instead run the validator of mntner:mntner.

    The list of referred objects can contain multiple entries, which means that
    the value must refer to one of these objects (e.g. tech-c can refer to
    role or person).

    If the references are strong, this reference is included in reference checks
    on updates, i.e. adding an object with a strong reference to another object
    that does not exist, is a validation failure.
    """
    keep_case = False

    def __init__(self, referring: List[str], strong=True, *args, **kwargs) -> None:
        from .parser import RPSLObject
        self.referring = referring
        self.strong = strong
        self.referring_object_classes: List[Type[RPSLObject]] = []
        self.referring_identifier_fields: List[RPSLTextField] = []
        super().__init__(*args, **kwargs)

    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if not self.referring_identifier_fields:
            self._build_cache()

        referring_field_messages = RPSLParserMessages()
        for identifier_field in self.referring_identifier_fields:
            if identifier_field:
                parsed_value = identifier_field.parse(value, referring_field_messages, strict_validation)
                if parsed_value is not None:
                    return parsed_value

        messages.merge_messages(referring_field_messages)
        return None

    def _build_cache(self):
        from .rpsl_objects import OBJECT_CLASS_MAPPING
        for ref in self.referring:
            rpsl_object_class = OBJECT_CLASS_MAPPING[ref]
            pk_field = [field for field in rpsl_object_class.fields.values() if field.primary_key and field.lookup_key][0]
            self.referring_object_classes.append(rpsl_object_class)
            self.referring_identifier_fields.append(pk_field)


class RPSLReferenceListField(RPSLFieldListMixin, RPSLReferenceField):
    """
    Field for a comma-seperated list of references to another field.

    Optionally, ANY can be allowed as a valid option too, instead of a list.
    """
    def __init__(self, allow_kw_any: bool=False, *args, **kwargs) -> None:
        self.allow_kw_any = allow_kw_any
        super().__init__(*args, **kwargs)

    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        if self.allow_kw_any and value.upper() == 'ANY':
            return RPSLFieldParseResult('ANY', values_list=['ANY'])
        return super().parse(value, messages, strict_validation)


class RPSLAuthField(RPSLTextField):
    """Field for the auth attribute of a mntner."""
    def parse(self, value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
        valid_beginnings = [hasher + ' ' for hasher in PASSWORD_HASHERS.keys()]
        has_valid_beginning = any(value.upper().startswith(b) for b in valid_beginnings)
        is_valid_hash = has_valid_beginning and value.count(' ') == 1 and not value.count(',')
        if is_valid_hash or re_pgpkey.match(value.upper()):
            return RPSLFieldParseResult(value)

        hashers = ', '.join(PASSWORD_HASHERS.keys())
        messages.error(f'Invalid auth attribute: {value}: supported options are {hashers} and PGPKEY-xxxxxxxx')
        return None


def parse_set_name(prefixes: List[str], value: str, messages: RPSLParserMessages, strict_validation=True) -> Optional[RPSLFieldParseResult]:
    assert all([prefix in reserved_prefixes for prefix in prefixes])
    input_components = value.split(':')
    output_components: List[str] = []
    prefix_display = '/'.join(prefixes)

    if strict_validation and len(input_components) > 5:
        messages.error('Set names can have a maximum of five components.')
        return None

    if strict_validation and not any([c.upper().startswith(tuple(prefixes)) for c in input_components]):
        messages.error(f'Invalid set name {value}: at least one component must be '
                       f'an actual set name (i.e. start with {prefix_display})')
        return None

    for component in input_components:
        if strict_validation and component.upper() in reserved_words:
            messages.error(f'Invalid set name {value}: component {component} is a reserved word')
            return None

        parsed_as_number = None
        try:
            parsed_as_number, _ = parse_as_number(component)
        except ValidationError:
            pass
        if not re_generic_name.match(component.upper()) and not parsed_as_number:
            messages.error(
                f'Invalid set {value}: component {component} is not a valid AS number nor a valid set name'
            )
            return None
        if strict_validation and not parsed_as_number and not component.upper().startswith(tuple(prefixes)):
            messages.error(f'Invalid set {value}: component {component} is not a valid AS number, '
                           f'nor does it start with {prefix_display}')
            return None

        if parsed_as_number:
            output_components.append(parsed_as_number)
        else:
            output_components.append(component)

    parsed_value = ':'.join(output_components)
    if parsed_value != value:
        messages.info(f'Set name {value} was reformatted as {parsed_value}')
    return RPSLFieldParseResult(parsed_value)


def clean_ip_value_error(value_error):
    return re.sub(re_clean_ip_error, '', str(value_error))



#!/usr/bin/env python
# flake8: noqa: E402

import sys

import argparse
import logging
from pathlib import Path

"""
Submit a raw e-mail message, i.e. with e-mail headers.
The message is always read from stdin.

A report on the results will be sent to the user by e-mail.
"""

logger = logging.getLogger(__name__)
sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.conf import config_init, CONFIG_PATH_DEFAULT
from irrd.updates.email import handle_email_submission


def run(data):
    try:
        handle_email_submission(data)
    except Exception as exc:
        logger.critical(f'An exception occurred while attempting to process the following email: {data}', exc_info=exc)
        print('An internal error occurred while processing this email.')


def main():  # pragma: no cover
    description = """Process a raw email message with requested changes. Authentication is checked, message
                     is always read from stdin. A report is sent to the user by email, along with any
                     notifications to mntners and others."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    args = parser.parse_args()

    config_init(args.config_file_path)

    run(sys.stdin.read())


if __name__ == '__main__':  # pragma: no cover
    main()
